from typing import Tuple

from sentio_prober_control.Sentio.Enumerations import ThermoChuckState
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.ModuleCommandGroupBase import (
    ModuleCommandGroupBase,
)


class StatusCommandGroup(ModuleCommandGroupBase):
    """A command group for getting the status of the probe station and controlling the dashboard module."""

    def __init__(self, comm) -> None:
        super().__init__(comm, "status")


    def get_chuck_temp(self) -> float:
        """Get current chuck temperature.

        Returns:
            The chuck temperature in degrees Celsius.
        """

        self.comm.send("status:get_chuck_temp")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        temp = float(tok[0])
        return temp


    def get_chuck_temp_setpoint(self) -> float:
        """Get current chuck temperature setpoint.

        Returns:
            The chuck temperature setpoint in degrees Celsius.
        """

        self.comm.send("status:get_chuck_temp_setpoint")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        temp = float(tok[0])
        return temp


    def get_chuck_thermo_state(self) -> ThermoChuckState:
        """Return thermo chuck state.

        Returns:
            A tuple with the current state of the thermo chuck. Contains six boolean variables:
            isCooling, isHeating, isControlling, isStandby, isError, isUncontrolled.
        """

        self.comm.send("status:get_chuck_thermo_state")
        resp = Response.check_resp(self.comm.read_line())

        if "soaking" in resp.message().lower():
            return ThermoChuckState.Soaking
        elif "cooling" in resp.message().lower():
            return ThermoChuckState.Cooling
        elif "heating" in resp.message().lower():
            return ThermoChuckState.Heating
        elif "controlling" in resp.message().lower():
            return ThermoChuckState.Controlling
        elif "standby" in resp.message().lower():
            return ThermoChuckState.Standby
        elif "error" in resp.message().lower():
            return ThermoChuckState.Error
        elif "uncontrolled" in resp.message().lower():
            return ThermoChuckState.Uncontrolled
        else:
            return ThermoChuckState.Unknown


    def get_machine_status(self) -> Tuple[bool, bool, bool]:
        """Get machine status.

        Returns:
            A tuple with the current status of the machine. Contains three boolean variables: isInitialized, isMeasuring, LoaderBusy.
        """

        self.comm.send("status:get_machine_status")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        isInitialized = "Ready" in tok
        isMeasuring = "IsMeasuring" in tok
        LoaderBusy = "LoaderBusy" in tok
        return isInitialized, isMeasuring, LoaderBusy


    def get_soaking_time(self, temperature: float):
        """Get the thermochuck soaking time in seconds that is set up for a certain temperature in the dashboard.

        Args:
            temperature: The temperature. This temperature value must be one of the predefined temperature
                  values set up in the dashboard.

        Raises:
            ProberException: With error code 200 is thrown when the temperature is not a predefined value set up in the dashboard.
            May also raise an an exception when other errors occur.

        Returns:
            The soaking time in seconds.
        """

        self.comm.send(f"status:get_soaking_time {temperature:.2f}")
        resp = Response.check_resp(self.comm.read_line())
        temp = float(resp.message())
        return temp


    def set_chuck_temp(self, temp: float) -> Response:
        """Set chuck temperature setpoint.

        Args:
            temp: The chuck temperature setpoint in degrees Celsius.
        """

        self.comm.send(f"status:set_chuck_temp {temp:.2f}")
        return Response.check_resp(self.comm.read_line())



""" # The Sentio package

    The Sentio package is your gateway to control a probe station running the MPI SENTIO Software suite. 
    To start you need two things: A communication channel and a `SentioProber` instance.

    ## Creating a Prober Instance

    The following example will Connect to a probe station and switch to SENTIO's wafermap module.

    >>> from sentio_prober_control.Sentio.ProberSentio import SentioProber
    >>> from sentio_prober_control.Communication.CommunicatorTcpIp import CommunicatorTcpIp
    >>> prober = SentioProber(CommunicatorTcpIp.create("127.0.0.1:35555"))
    >>> prober.select_module(Module.Wafermap)
"""
print(f"Initializing package {__name__}")



from typing import Tuple

from sentio_prober_control.Sentio.Enumerations import RoutingPriority, RoutingStartPoint, TestSelection
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase


class WafermapPathCommandGroup(CommandGroupBase):
    """This command group bundles functions for setting up and using the test path of the wafermap.

    A test path defines which dies are tested in which order.
    """

    def create_from_bin(self, bin_val: int) -> None:
        """Create test path by using all dies with a specific bin.

        Wraps SENTIO's map:path:create_from_bin remote command.

        Args:
            bin_val: The bin value to use.
        """
        self.comm.send("map:path:create_from_bins {0}".format(bin_val))
        Response.check_resp(self.comm.read_line())


    def get_die(self, seq: int) -> Tuple[int, int]:
        """Get die column and row coordinates from a sequence number.

        Wraps SENTIO's map:path:get_die remote command.

        Args:
            seq: The sequence number of the die.

        Returns:
            A tuple with the column and row coordinates of the die.
        """
        self.comm.send("map:path:get_die {0}".format(seq))
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return int(tok[0]), int(tok[1])


    def select_dies(self, selection: TestSelection) -> None:
        """Select dies for testing.

        Wraps SENTIO's map:path:select_dies remote command.

        Args:
            selection: The selection of dies to select.
        """
        self.comm.send(f"map:path:select_dies {selection.toSentioAbbr()}")
        Response.check_resp(self.comm.read_line())


    def set_routing(self, sp: RoutingStartPoint, pri: RoutingPriority) -> None:
        """Set up path finnding for stepping by specifying a start point position
        and a row or column priority for routing.

        Wraps SENTIO's map:set_routing remote command.

        Args:
            sp: The start point of the routing.
            pri: The priority of the routing (rows first, columns first).
        """
        self.comm.send(f"map:set_routing {sp.toSentioAbbr()}, {pri.toSentioAbbr()}")
        Response.check_resp(self.comm.read_line())



from deprecated import deprecated

from sentio_prober_control.Sentio.ProberBase import ProberException


class Response:
    """This class represents the response of a single SENTIO remote command.

    Sentio's remote command response is a comma separated string that contains three fields
    which represent 4 data items. (error code and status code are combined)

    """

    def __init__(self, errc: int, stat: int, cmd_id: int, msg: str):
        """Creates a new Response object.

        Args:
            errc (int): The error code.
            stat (int): The status code.
            cmd_id (int): The async command id. (only used by async commands)
            msg (str): The response message.
        """
        self.__errc = errc
        self.__stat = stat
        self.__cmd_id = cmd_id
        self.__msg = msg


    @staticmethod
    def parse_resp(resp):
        """A static method that parses a SENTIO remote command response string and returns a Response object.

        A typical response from SENTIO to a remote command might look like "0,0,ok"
        SENTIO's remote command responses are strings that contain multiple items
        separated by two commas.

        - error code and status information (combined in one integer)
        - an async command id (only used by async commands)
        - a response message

        Returns:
            response (Resppnse): A Response object created from the information in SENTIO's response string.
        """
        tok = resp.split(",", 2)

        # split response items
        errc = int(tok[0]) & 1023  # lowermost 10 bits are the error code
        stat = (int(tok[0]) & ~1023) >> 10  # everything from bit 10 on is the status

        cmd_id = int(tok[1])
        msg = tok[2].rstrip()

        # seperate error code an status code
        resp = Response(errc, stat, cmd_id, msg)
        # resp.dump()
        return resp
    

    @staticmethod
    def check_resp(str_resp: str) -> "Response":
        """A static method that parses a response string and raises an exception if the response indicates an error.

        Args:
            str_resp (str): The response string to parse.

        Returns:
            response (Response): A Response object created from the information in SENTIO's response string. If the response represents
                                 an error an exception is raised instead of returning a Response object.

        Raises:
            ProberException: If the response indicates an error.
        """
        resp = Response.parse_resp(str_resp)
        if not resp.ok():
            raise ProberException(resp.message(), resp.errc())

        return resp


    def check(self):
        """Raises an exception if this response indicates an error.

        Raises:
            ProberException: If the response indicates an error.
        """
        if not self.ok():
            raise ProberException(self.message(), self.errc())


    def cmd_id(self) -> int:
        """The async commans id returned by SENTIO.

        If the remote command is not an async command 0 is returned.

        Returns:
            cmd_id (int): The async command id returned by SENTIO.
        """
        return self.__cmd_id


    def errc(self) -> int:
        """The error code returned by SENTIO.

        The meaning of the error code is documented in the SENTIO's remote command documentation.
        It is also used by the enumerator RemoteCommandError.

        Returns:
            errc (int): The error code returned by SENTIO.
        """
        return self.__errc


    def message(self) -> str:
        """The response message returned by SENTIO.

        Returns:
            msg (str): The response message returned by SENTIO.
        """
        return self.__msg


    def status(self):
        """The status coode extracted from the response.

        Returns:
            stat (int): The status code returned by SENTIO.
        """
        return self.__stat


    def ok(self) -> bool:
        """Returns True if the response indicates no error.

        Returns:
            ok (bool): True if the response indicates no error, False otherwise.
        """
        return self.__errc == 0


    @deprecated("Use print() instead")
    def dump(self) -> None:
        print('errc={0}; stat={1}; msg="{2}"; id={3}'.format(self.__errc, self.__stat, self.__msg, self.__cmd_id))


    def print(self) -> None:
        """Prints the content of the response object to the console."""
        print('errc={0}; stat={1}; msg="{2}"; id={3}'.format(self.__errc, self.__stat, self.__msg, self.__cmd_id))



import base64
from typing import Tuple

from deprecated import deprecated

from sentio_prober_control.Communication.CommunicatorBase import CommunicatorBase
from sentio_prober_control.Sentio.Enumerations import (
    AutoAlignCmd,
    AutoFocusCmd,
    CameraMountPoint,
    DetectionAlgorithm,
    DetectionCoordindates,
    DieCompensationMode,
    DieCompensationType,
    FindPatternReference,
    PtpaFindTipsMode,
    PtpaType,
    SnapshotLocation,
    SnapshotType,
)
from sentio_prober_control.Sentio.ProberBase import ProberException
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.ModuleCommandGroupBase import ModuleCommandGroupBase
from sentio_prober_control.Sentio.CommandGroups.VisionCameraCommandGroup import VisionCameraCommandGroup
from sentio_prober_control.Sentio.CommandGroups.VisionCompensationGroup import VisionCompensationGroup
from sentio_prober_control.Sentio.CommandGroups.VisionIMagProCommandGroup import VisionIMagProCommandGroup


class VisionCommandGroup(ModuleCommandGroupBase):
    """This command group contains functions for working with SENTIO's vision module.
    You are not meant to instantiate this class directly. Access it via the vision attribute
    of the [SentioProber](SentioProber.md) class.

    Attributes:
        camera (VisionCameraCommandGroup): A subgroup to provide logic for camera specific functions.
        imagpro (VisionIMagProCommandGroup): A subgroup to provide logic for IMagPro specific functions.
        compensation (VisionCompensationGroup): A subgroup to provide logic for compensation specific functions.
    """

    def __init__(self, comm: CommunicatorBase) -> None:
        super().__init__(comm, "vis")

        self.camera = VisionCameraCommandGroup(comm)
        self.imagpro = VisionIMagProCommandGroup(comm)
        self.compensation = VisionCompensationGroup(comm)


    def align_wafer(self, mode: AutoAlignCmd) -> Response:
        """Perform a wafer alignment.

        Args:
            mode: The alignment procedure to use.

        Returns:
            A Response object.
        """

        self.comm.send(f"vis:align_wafer {mode.toSentioAbbr()}")
        return Response.check_resp(self.comm.read_line())


    def align_die(self, threshold: float = 0.05) -> Tuple[float, float, float]:
        """Perform a die alignment.

        Die alignment compensates for positional differences when working with diced wafers
        which contain single dies.

        Args:
            threshold: The alignment threshold.

        Returns:
            A tuple with the x, y and theta offset in micrometer.
        """

        self.comm.send(f"vis:align_die {threshold}")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1]), float(tok[2])


    def auto_focus(self, af_cmd: AutoFocusCmd = AutoFocusCmd.Focus) -> float:
        """Perform an auto focus operation.

        Args:
            af_cmd: The auto focus command to execute.

        Returns:
            The focus height in micrometer
        """

        resp = self.prober.send_cmd(f"vis:auto_focus {af_cmd.toSentioAbbr()}")
        tok = resp.message().split(",")
        return float(tok[0])


    def camera_synchronize(self) -> Tuple[float, float, float]:
        self.comm.send("vis:camera_synchronize")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1]), float(tok[2])


    def create_probepad_model(self, angleStep: float = 0.1, imgPath: str | None = None, UL: tuple | None = None, LR: tuple | None = None):
        if imgPath == None or UL == None or LR == None:
            self.comm.send(f"vis:create_probepad_model {angleStep}")
        else:
            self.comm.send(f"vis:create_probepad_model {angleStep}, {imgPath}, {UL[0]}, {UL[1]}, {LR[0]}, {LR[1]}")

        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return tok


    def detect_probepads(self, imgPath: str | None = None, minScore: float = 0.7, startAngle: float | None = None, startExtend: float | None = None, maxOverlap: float | None  = None) -> Tuple[bool, float, str]:
        if startAngle == None:
            self.comm.send("vis:detect_probepads {},{}".format(imgPath, minScore))
        else:
            self.comm.send("vis:detect_probepads {},{},{},{},{}".format(imgPath, minScore, startAngle, startExtend, maxOverlap))

        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return bool(tok[0]), float(tok[1]), tok[2]


    def detect_probetips(self, camera: CameraMountPoint, detector: DetectionAlgorithm = DetectionAlgorithm.ProbeDetector, coords: DetectionCoordindates = DetectionCoordindates.Roi) -> list:
        """Executes a built in detector on a given camera and return a list of detection results.

        Args:
            camera: The camera to use for detection.
            detector: The detection algorithm to use.
            coords: The coordinates to use for the returned detection results.

        Returns:
            A list of detected tips. Each detection result is a tuply of 6 values: x, y, width, height, score, class_id.
        """

        self.comm.send(f"vis:detect_probetips {camera.toSentioAbbr()}, {detector.toSentioAbbr()}, {coords.toSentioAbbr()}")
        resp = Response.check_resp(self.comm.read_line())
        str_tips = resp.message().split(",")

        found_tips = []
        cid : float = 0
        for n in range(0, len(str_tips)):
            str_tip = str_tips[n].strip().split(" ")
            num_col = len(str_tip)

            x = float(str_tip[0])  # tip x position
            y = float(str_tip[1])  # tip y position
            w = float(str_tip[2])  # detection width
            h = float(str_tip[3])  # detection height
            q = float(str_tip[4])  # detection quality (meaning depends on the used detector)

            if num_col >= 6:
                cid = float(str_tip[5])  # class id (only multi class detectors)

            found_tips.append([x, y, w, h, q, cid])

        return found_tips


    def enable_follow_mode(self, stat: bool) -> Response:
        """Enable or disable the scope follow mode.

        If scope follow mode is active the scope will move in sync with the chuck. This is useful for
        keeping the image in focus while moving the chuck.

        This function wraps the "vis:enable_follow_mode" remote command.

        Args:
            stat: A flag indicating whether to enable or disable the follow mode.

        Returns:
            A Response object.
        """

        self.comm.send("vis:enable_follow_mode {0}".format(stat))
        return Response.check_resp(self.comm.read_line())


    def find_home(self) -> Response:
        """Find home position.

        This function uses a pre-trained pattern to fully automatically find the home position.

        Returns:
            A Response object.
        """

        self.comm.send("vis:find_home")
        return Response.check_resp(self.comm.read_line())


    def find_pattern(self, name: str, threshold: float = 70, pattern_index: int = 0, reference: FindPatternReference = FindPatternReference.CenterOfRoi) -> Tuple[float, float, float, float]:
        """Find a trained pattern in the camera image.

        Args:
            name: The name of the pattern to find.
            threshold: The detection threshold. The higher the threshold, the more certain the detection must be.
            pattern_index: The index of the pattern to find. In SENTIO each pattern may have up to 5 alternate patterns. This is the index of the alternate pattern.
            reference: The reference point to use for the pattern detection.
        """

        self.comm.send(f"vis:find_pattern {name}, {threshold}, {pattern_index}, {reference.toSentioAbbr()}")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1]), float(tok[2]), float(tok[3])


    def has_camera(self, camera: CameraMountPoint) -> bool:
        """Check wether a given camera is present in the system.

        This function wraps the "vis:has_camera" remote command.

        Args:
            camera: The camera mount point to check.

        Returns:
            True if the camera is present, False otherwise.
        """

        self.comm.send(f"vis:has_camera {camera.toSentioAbbr()}")
        resp = Response.check_resp(self.comm.read_line())
        return resp.message().upper() == "1"


    def switch_all_lights(self, stat: bool) -> Response:
        """Switch all camera lights on or off.

        This function wraps the "vis:switch_all_lights" remote command.

        Args:
            stat: A flag indicating whether to switch the lights on or off.

        Returns:
            A Response object.
        """

        self.comm.send("vis:switch_all_lights {0}".format(stat))
        return Response.check_resp(self.comm.read_line())


    def remove_probetip_marker(self) -> Response:
        """Remove probetip marker from the camera display.

        Returns:
            A Response object.
        """

        self.comm.send("vis:remove_probetip_marker")
        return Response.check_resp(self.comm.read_line())


    def match_tips(self, ptpa_type: PtpaType) -> Tuple[float, float]:
        """For internal use only!
        This function is subject to change without any prior warning. MPI will not maintain backwards
        compatibility or provide support."""

        self.comm.send("vis:match_tips {0}".format(ptpa_type.toSentioAbbr()))
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1])


    def snap_image(self, file: str, what: SnapshotType = SnapshotType.CameraRaw, where: SnapshotLocation = SnapshotLocation.Prober) -> None:
        """Save a snapshot of the current camera image to a file.

        Args:
            file: The file name to save the image to.
            what: The type of snapshot to take.
            where: The location where to store the snapshot. By default this is the prober control computer. If SnapshotLocation.Local
                   is specified the image is download from the probe computer and stored loacally.

        Returns:
            A Response object.
        """

        if where == SnapshotLocation.Local:
            self.comm.send(f"vis:snap_image **download**, {what.toSentioAbbr()}")
            resp = Response.check_resp(self.comm.read_line())
            jpeg_data = base64.b64decode(resp.message())

            # Save the file locally
            with open(file, "wb") as f:
                f.write(jpeg_data)
        else:
            self.comm.send(f"vis:snap_image {file}, {what.toSentioAbbr()}")
            Response.check_resp(self.comm.read_line())


    def switch_light(self, camera: CameraMountPoint, stat: bool) -> Response:
        """Switch the light of a given camera on or off.

        Args:
            camera: The camera to switch the light for.
            stat: A flag indicating whether to switch the light on or off.

        Returns:
            A Response object.
        """

        self.comm.send(f"vis:switch_light {camera.toSentioAbbr()}, {stat}")
        return Response.check_resp(self.comm.read_line())


    def switch_camera(self, camera: CameraMountPoint) -> Response:
        """Switch the camera to use for the vision module.

        Args:
            camera: The camera to switch to.

        Returns:
            A Response object.
        """

        self.comm.send(f"vis:switch_camera {camera.toSentioAbbr()}")
        return Response.check_resp(self.comm.read_line())


    def ptpa_find_pads(self, row: int = 0, column: int = 0):
        self.comm.send("vis:execute_ptpa_find_pads {0},{1}".format(row, column))
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1]), float(tok[2])


    def ptpa_find_tips(self, ptpa_mode: PtpaFindTipsMode):
        self.comm.send("vis:ptpa_find_tips {0}".format(ptpa_mode.toSentioAbbr()))
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1]), float(tok[2])


    def start_fast_track(self) -> Response:
        """Start the fast track process as defined in SENTIO.

        FastTrack is a feature that combines multiple preparation steps into one command.
        You need to have set up FastTrack in the VisionModule/Automation/FastTrack side 
        panel editor to use this command.
        
        Returns:
            A Response object.
        """

        return self.prober.send_cmd("vis:start_fast_track")


    @deprecated("use vision.compensation.start_execute(...) instead!")
    def start_execute_compensation(self, comp_type: DieCompensationType, comp_mode: DieCompensationMode):
        self.comm.send("vis:compensation:start_execute {0},{1}".format(comp_type.toSentioAbbr(), comp_mode.toSentioAbbr()))
        resp = Response.check_resp(self.comm.read_line())

        if not resp.ok():
            raise ProberException(resp.message())

        return resp.cmd_id()



from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase


class WafermapDieCommandGroup(CommandGroupBase):
    """This Command group bundles commands for setting up dies on a wafermap."""

    def add(self, x: int, y: int) -> None:
        """Add a die to the wafermap. If the die is already part of the map
        nothing happens.

        Args:
            x: The column of the die.
            y: The row of the die.
        """
        self.comm.send(f"map:die:add {x}, {y}")
        Response.check_resp(self.comm.read_line())


    def remove(self, x: int, y: int) -> None:
        """Remove a die from the wafermap.

        This will mark the die as nonexistant and make it unavailable for stepping.
        Removed dies are treated as if they were physically not present on the wafer.

        Wraps SENTIO's map:die:remove remote command.

        Args:
            x: The column of the die.
            y: The row of the die.
        """
        self.comm.send(f"map:die:remove {x}, {y}")
        Response.check_resp(self.comm.read_line())


    def select(self, x: int, y: int) -> None:
        """Selects a die for testing by adding it to the test path.

        Args:
            x: The column of the die.
            y: The row of the die.
        """
        self.comm.send(f"map:die:add {x}, {y}")
        Response.check_resp(self.comm.read_line())


    def unselect(self, x: int, y: int) -> None:
        """Unselects a die from testing by removing it from the test path.

        Args:
            x: The column of the die.
            y: The row of the die.
        """
        self.comm.send(f"map:die:unselect {x}, {y}")
        Response.check_resp(self.comm.read_line())



"""  
# SENTIO® Prober Control - Python Bindings
This archive contains a package with python bindings to control a [MPI SENTIO® probe station](https://www.mpi-corporation.com/ast/engineering-probe-systems/mpi-sentio-software-suite/).

![AST_Back_2A fw_](https://user-images.githubusercontent.com/2202567/204108957-0c7a864a-a526-4d32-a1ca-51985a0b01c6.png)

## Installation

The package for controlling MPI probe stations running the MPI SENTIO Software suite is available via [pythons package index](https://pypi.org/project/sentio-prober-control/). 
To install simply type:

```python -m pip install sentio-prober-control```

## Example-Scripts

A set of example scripts for python is maintained in a separate archive at GitHub. 

https://github.com/SentioProberDev/Examples-Python

"""
# from .Sentio.ProberSentio import SentioProber, Module
# from .Communication.CommunicatorTcpIp import CommunicatorTcpIp
# from .Communication.CommunicatorGpib import CommunicatorGpib

print(f"Initializing package {__name__}")

name = "sentio_prober_control"
""" The name of the package. 
    @private
"""



from typing import Tuple

from sentio_prober_control.Sentio.Enumerations import IMagProZReference
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase


class VisionIMagProCommandGroup(CommandGroupBase):
    """This command group contains functions for working with an IMag pro microscope.

    You are not meant to instantiate this class directly. Access it via the imagpro attribute
    of the vision attribute of the [SentioProber](SentioProber.md) class.
    """

    def __init__(self, comm) -> None:
        super().__init__(comm)


    def move_z(self, ref: IMagProZReference, pos: float) -> float:
        """Move imagpro's internal z-axis

        Args:
            ref: The position reference for the motion command.
            pos: The position to move to.

        Returns:
            The actual position of the z-axis after the move command.
        """

        #
        # the imag pro axis is showing hysteresis behavior. z position seems to be only
        # reproducable when the axis is moved from one of its endpoints! If you remove
        # the next three lines. The z-position will become very inaccurate!
        #
        # self.comm.send("vis:imagpro:move_z center, -1000")
        # Response.check_resp(self.comm.read_line())
        # time.sleep(0.5)

        self.comm.send("vis:imagpro:move_z {0}, {1}".format(ref.toSentioAbbr(), pos))
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())


    def get_z(self, ref: IMagProZReference) -> float:
        """Get the position od imagpro's internal axis.

        Args:
            ref: The position reference for the returned value.

        Returns:
            The position of the z-axis.
        """

        par: str = ref.toSentioAbbr()
        self.comm.send(f"vis:imagpro:get_z {par}")
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())


    def get_xy_comp(self, imag_pro_z: float) -> Tuple[float, float]:
        """Get the xy compensation value for a certain z-position of imagpro's internal axis

        Args:
            imag_pro_z: The z-position of imagpro's internal axis.

        Returns:
            The xy compensation value
        """

        self.comm.send(f"vis:imagpro:get_xy_comp {imag_pro_z}")

        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1])



import re

from typing import List


class Helper:
    def split_string(s : str, delimiter : str) -> List[str]:
        escaped_delimiter = re.escape(delimiter)
        pattern = rf'"[^"]*"|[^{escaped_delimiter}]+'
        matches = re.findall(pattern, s)

        return [match.strip('"') for match in matches]


from deprecated import deprecated

from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase


class QAlibriaCommandGroup(CommandGroupBase):
    """This command group contains functions for working with Qalibria."""

    def __init__(self, comm) -> None:
        super().__init__(comm)


    def calibration_execute(self) -> None:
        """Execute VNA calibration.

        Wraps SENTIO's "qal:calibration_execute" remote command.
        """
        self.comm.send("qal:calibration_execute")
        Response.check_resp(self.comm.read_line())


    def calibration_drift_verify(self, dut_name: str = "", auto_exec: bool = True) -> None:
        """Calibration Drift verification.

        Wraps SENTIO's "qal:calibration_drift_verify" remote command.

        Args:
            dut_name: The name of the dut.
            auto_exec:
        """
        self.comm.send("qal:calibration_drift_verify {dut_name}, {auto_exec}")
        resp = Response.check_resp(self.comm.read_line())

    # Inconsistent API filed as CR#13887. This function should not exist!
    # the use of the "start_" prefix implies it is an async function. It is not or
    # if it is where it the command id return value?
    @deprecated(reason="use calibration_execute instead!; this function violates naming conventions since its name is different from the remote command; filed as CR#13887")
    def start_calibration(self) -> None:
        self.comm.send("qal:calibration_execute")
        Response.check_resp(self.comm.read_line())


    # Inconsistent API filed as CR#13887. This function should not exist!
    @deprecated(reason="use calibration_drift_verify instead!; this function violates naming conventions since its name is different from the remote command; filed as CR#13887")
    def verify_calibration_drift(self) -> None:
        self.comm.send("qal:calibration_drift_verify")
        Response.check_resp(self.comm.read_line())


    # Inconsistent API filed as CR#13887. This function should not exist!
    @deprecated(reason="use calibration_drift_verify instead!; this function violates naming conventions since its name is different from the remote command; filed as CR#13887")
    def verify_calibration_drift_dut(self, dut) -> None:
        self.comm.send("qal:calibration_drift_verify {}".format(dut))
        Response.check_resp(self.comm.read_line())


    # Inconsistent API filed as CR#13887. This function should not exist!
    @deprecated(reason="oddly specific function name; filed as CR#13887")
    def set_calibration_drift_probe12(self):
        self.comm.send(" qal:set_dut_network RefDUT,DriftRef,12,false")
        resp = Response.check_resp(self.comm.read_line())

        self.comm.send(" qal:set_dut_network RefDUT,Drift,12,false")
        resp = Response.check_resp(self.comm.read_line())
        return resp.message()



import pyvisa

from sentio_prober_control.Communication.CommunicatorBase import CommunicatorBase


class CommunicatorVisa(CommunicatorBase):
    """Communicator for VISA communication.

    This communicator is meant for communication over National Instruments VISA interface.
    In order to use it you must have the [NI-VISA driver](https://www.ni.com/de/support/downloads/drivers/download.ni-visa.html) installed on your system.
    """

    def __init__(self):
        """Constructs a VISA communicator."""
        self.__rm = pyvisa.ResourceManager()

    @staticmethod
    def create(addr: str):
        """Create an instance of a VISA communicator.

        :param addr: A string that specifies the address of the VISA device to connect to. The address must be a valid VISA resource identifier.
        """
        c = CommunicatorVisa()
        c.connect(addr)
        return c

    def connect(self, address: str):
        """Connect to a VISA device at the specified address.

        The VISA timeout for the interface is set to 3600000 (1 hour).

        :param addr: A string that specifies the address of the VISA device to connect to. The address must be a valid VISA resource identifier.
        """
        self.__visa = self.__rm.open_resource(address)
        self.__visa.timeout = 3600000

    def disconnect(self):
        """Disconnect from the VISA device."""
        if CommunicatorBase._verbose:
            print("Diconnecting comunicator to {0}".format(self.__address))

        self.__visa.close()

    def send(self, msg: str):
        """Send a command to the VISA device.

        :param msg: The command to send.
        """
        if CommunicatorBase._verbose:
            print('Sending "{0}"'.format(msg))

        self.__visa.write((msg + "\n"))

    def read_line(self):
        """Read a line from the VISA device.

        :return: The read line.
        """
        return self.__visa.read()



from deprecated import deprecated

from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.ProberBase import ProberException
from sentio_prober_control.Sentio.CommandGroups.AuxCleaningGroup import AuxCleaningGroup
from sentio_prober_control.Sentio.CommandGroups.ModuleCommandGroupBase import (
    ModuleCommandGroupBase,
)


class AuxCommandGroup(ModuleCommandGroupBase):
    """This command group contains functions for working with auxiliary sites of the chuck.
    You are not meant to create instances of this class on your own. Instead use the aux property
    of the SentioProber class.

    Attributes:
        cleaning (AuxCleaningGroup): A subgroup to provide logic for probe cleaning.
    """

    def __init__(self, comm) -> None:
        super().__init__(comm, "aux")

        self.cleaning : AuxCleaningGroup = AuxCleaningGroup(comm)





from enum import Enum

from deprecated import deprecated


class AutoAlignCmd(Enum):
    """Specifies an algorithm for performaing auto alignment.

    Attributes:
        AutoDieSize (0): Perform auto alignment. Determine die size with zero prior knowledge.
        UpdateDieSize (1): Perform auto alignment. Update die size based on current value.
        TwoPt (2): Use Two-Point alignment algorithm.
    """

    AutoDieSize = 0
    UpdateDieSize = 1
    TwoPt = 2

    def toSentioAbbr(self):
        switcher = {
            AutoAlignCmd.AutoDieSize: "auto",
            AutoAlignCmd.UpdateDieSize: "update",
            AutoAlignCmd.TwoPt: "2pt",
        }
        return switcher.get(self, "Invalid Auto Align function")


class AutoFocusAlgorithm(Enum):
    """A list of different AutoFocus algorithms.

    Attributes:
        Gradient (0): Gradient Engery
        Bandpass (1): Bandpass Filter
        Difference (2): Difference
        AutoCorrelation (3): Auto Correlation
    """

    Gradient = 0
    Bandpass = 1
    Difference = 2
    AutoCorrelation = 3

    def toSentioAbbr(self):
        switcher = {
            AutoFocusAlgorithm.Gradient: "Gradient",
            AutoFocusAlgorithm.Bandpass: "Bandpass",
            AutoFocusAlgorithm.Difference: "Difference",
            AutoFocusAlgorithm.AutoCorrelation: "AutoCorrelation",
        }
        return switcher.get(self, "Invalid focus measure")


class AutoFocusCmd(Enum):
    """A list of different AutoFocus functions.

    Attributes:
        Calibration (0): Execute focus calibration.
        Focus (1): Measure Focus curve and determine maximum.
        GoTo (2): Simply move the scope to distance from the wafer that is known to be in focus.
    """

    Calibration = 0
    Focus = 1
    GoTo = 2

    def toSentioAbbr(self):
        switcher = {
            AutoFocusCmd.Calibration: "C",
            AutoFocusCmd.Focus: "F",
            AutoFocusCmd.GoTo: "G",
        }
        return switcher.get(self, "Invalid auto focus function")


class AxisOrient(Enum):
    """Represents an axis orientation.

    Thie enumeration is mostly used to describe axis orientations
    on the computer screen. For instance the axis orientation
    of the wafer map.

    Attributes:
        DownRight (0): Y axis pointing down; X axis pointing right.
        DownLeft (1): Y axis pointing down; X axis pointing left.
        UpRight (2): Y-axis pointing up; X axis pointing right.
        UpLeft (3): Y-axis pointing up; X axis pointing left.
    """

    DownRight = 0
    DownLeft = 1
    UpRight = 2
    UpLeft = 3

    def toSentioAbbr(self):
        switcher = {
            AxisOrient.DownRight: "DR",
            AxisOrient.DownLeft: "DL",
            AxisOrient.UpRight: "UR",
            AxisOrient.UpLeft: "UL",
        }
        return switcher.get(self, "Invalid AxisOrient")


class BinSelection(Enum):
    """An enumerator for selecting dies for binning.

    Attributes:
        All (0): Select dies and subsites.
        DiesOnly (1): Select dies only.
        SubsitesOnly (2): Select subsites only.
    """

    All = 0
    DiesOnly = 1
    SubsitesOnly = 2

    def toSentioAbbr(self):
        switcher = {
            BinSelection.All: "a",
            BinSelection.DiesOnly: "d",
            BinSelection.SubsitesOnly: "s",
        }
        return switcher.get(self, "Invalid bin selection")


class BinQuality(Enum):
    """An enumerator dor defining the quality parameter of a certain bin code.

    Attributes:
        All (0): Select dies and subsites.
        DiesOnly (1): Select dies only.
        SubsitesOnly (2): Select subsites only.
    """

    Pass = 0
    Fail = 1
    Undefined = 2

    def toSentioAbbr(self):
        switcher = {
            BinQuality.Pass: 0,
            BinQuality.Fail: 1,
            BinQuality.Undefined: 2,
        }
        return switcher.get(self, "Invalid bin quality identifier")


class CameraMountPoint(Enum):
    """Available camera mount points.

    A camera mount point is a physical position in the prober where a camera can be located.
    SENTIO refers to its camera via the camera mount point.

    Attributes:
        Scope (0): The downward looking microscope camera.
        Scope2 (1): Second scope camera. This camera is only used by IMag to provide a wider field of view in addition to the scope camera.
        Chuck (2): The upward looking chuck camera.
        OffAxis (3): Downward looking platen camera.
        Vce (4): First Vce camera.
        Vce2 (5): Second Vce camera.
    """

    Scope = 0
    Scope2 = 1
    Chuck = 2
    OffAxis = 3
    Vce = 4
    Vce2 = 5

    def toSentioAbbr(self):
        switcher = {
            CameraMountPoint.Scope: "scope",
            CameraMountPoint.Scope2: "scope2",
            CameraMountPoint.Chuck: "chuck",
            CameraMountPoint.OffAxis: "offaxis",
            CameraMountPoint.Vce: "vce01",
            CameraMountPoint.Vce2: "vce02",
        }
        return switcher.get(self, "Invalid camera mount point id")


class ChuckPositionHint(Enum):
    """Position hint for the chuck stage.

    Attributes:
        Center (0): Chuck is at Center Position
        FrontLoad (1): Chuck is at Front Load Position
        SideLoad (2): Chuck is at Side Load Position
    """

    Center = 0
    FrontLoad = 1
    SideLoad = 2


class ChuckSite(Enum):
    """An enumeration containing chuck sites.

    A chuck site is a physical position that is attached to the
    chuck and is moving together with the chuck.

    Attributes:
        Wafer (0): The wafer site. This is where your wafer is supposed to be.
        AuxRight (1): Right auxilliary site (if available)
        AuxLeft (2): Left auxilliary site (if available)
        AuxRight2 (3): Secondary right auxilliary site (if available)
        AuxLeft2 (4): Secondary left auxilliary site (if available)
        ChuckCamera (5): The chuck camera
    """

    Wafer = 0
    AuxRight = 1
    AuxLeft = 2
    AuxRight2 = 3
    AuxLeft2 = 4
    ChuckCamera = 5

    def toSentioAbbr(self):
        switcher = {
            ChuckSite.Wafer: "Wafer",
            ChuckSite.AuxLeft: "AuxLeft",
            ChuckSite.AuxLeft2: "AuxLeft2",
            ChuckSite.AuxRight: "AuxRight",
            ChuckSite.AuxRight2: "AuxRight2",
            ChuckSite.ChuckCamera: "ChuckCamera",
        }
        return switcher.get(self, "Invalid chuck site")


class ChuckThetaReference(Enum):
    """Reference to use for chuck theat motions.

    Attributes:
        Zero (0): Use zero of the theta axis.
        Site (1): Use the trained site of "home" angle as the reference.
        Relative (2): Use the current theta position as reference.
    """

    Zero = 0
    Site = 1
    Relative = 2

    def toSentioAbbr(self):
        switcher = {
            ChuckThetaReference.Zero: "Z",
            ChuckThetaReference.Site: "S",
            ChuckThetaReference.Relative: "R",
        }
        return switcher.get(self, "Invalid chuck theta reference")


class ChuckXYReference(Enum):
    """Defines a reference for chuck xy motions.

    Attributes:
        Zero (0): Use absolute chuck coordinates.
        Home (1): Use home position as reference.
        Relative (2): Use curent chuck position as reference.
        Center (3): Use chuck center position as reference.
        User (4): Use user defined coordinate system.
    """

    Zero = 0
    Home = 1
    Relative = 2
    Center = 3
    User = 4

    def toSentioAbbr(self):
        switcher = {
            ChuckXYReference.Zero: "Z",
            ChuckXYReference.Home: "H",
            ChuckXYReference.Relative: "R",
            ChuckXYReference.Center: "C",
            ChuckXYReference.User: "U",
        }
        return switcher.get(self, "Invalid chuck xy reference")


class ChuckZReference(Enum):
    """Defines a position reference for chuck z-motions.

    Attributes:
        Zero (0): Use absolute chuck z coordinates with respect the the physical axis zero positon.
        Relative (1): Use relative chuck z coordinated with respect to the current position.
        Contact (2): Use relative chuck z coordinated with respect to the chucks contact height.
        Hover (3): Use relative chuck z coordinated with respect to the chucks hover height.
        Separation (4): Use relative chuck z coordinated with respect to the chucks separation height.
    """

    Zero = 0
    Relative = 1
    Contact = 2
    Hover = 3
    Separation = 4

    def toSentioAbbr(self):
        switcher = {
            ChuckZReference.Zero: "Z",
            ChuckZReference.Relative: "R",
            ChuckZReference.Contact: "C",
            ChuckZReference.Hover: "H",
            ChuckZReference.Separation: "S",
        }
        return switcher.get(self, "Invalid chuck z reference")


class ColorScheme(Enum):
    """The color scheme used by the wafer map.

    Attributes:
        ColorFromBin (0): The color of a die is determined by the bin code of the die.
        ColorFromValue (1): The color of the die is determined by a floating point value attached to a die.
    """

    ColorFromBin = 0
    ColorFromValue = 1

    def toSentioAbbr(self):
        switcher = {ColorScheme.ColorFromBin: 0, ColorScheme.ColorFromValue: 1}
        return switcher.get(self, "Invalid ColorScheme")


@deprecated(reason="duplicated; use DieCompensationMode instead.")
class Compensation(Enum):
    Lateral = 0
    Vertical = 1
    Both = 2
    ProbeCard = 3
    MapScan = 4
    Thermal = 5
    Topography = 6

    def toSentioAbbr(self):
        switcher = {
            Compensation.Lateral: "lateral",
            Compensation.Vertical: "vertical",
            Compensation.Both: "both",
            Compensation.ProbeCard: "probecard",
            Compensation.MapScan: "mapscan",
            Compensation.Thermal: "thermal",
            Compensation.Topography: "topography",
        }
        return switcher.get(self, "Invalid compensation type")


class CompensationMode(Enum):
    """A list with available compensation modes.

    Attributes:
        Lateral (0): Lateral (XY) compensation.
        Vertical (1): Vertical (Z) compensation.
        Both (2): Both lateral and vertical compensation.
        ProbeCard (3): Probe card compensation.
        MapScan (4): MapScan compensation.
        Topography (5): Topography compensation.
    """

    Lateral = 0
    Vertical = 1
    Both = 2
    ProbeCard = 3
    MapScan = 4
    Thermal = 5
    Topography = 6

    def toSentioAbbr(self):
        switcher = {
            CompensationMode.Lateral: "Lateral",
            CompensationMode.Vertical: "Vertical",
            CompensationMode.Both: "Both",
            CompensationMode.ProbeCard: "ProbeCard",
            CompensationMode.MapScan: "mapscan",
            CompensationMode.Thermal: "thermal",
            CompensationMode.Topography: "topography",
        }
        return switcher.get(self, "Invalid CompensationMode")


class CompensationType(Enum):
    """A list of compensation types.

    Attributes:
        DieAlign (0):
        Topography (1): Vertical (Z) compensation.
        MapScan (2): Both lateral and vertical compensation.
        AlignDie (3): Probe card compensation.
        SkateDetection (4): MapScan compensation.
    """

    DieAlign = 0
    Topography = 1
    MapScan = 2
    AlignDie = 3
    SkateDetection = 4
    OnTheFly = 5
    OffAxis = 6

    def toSentioAbbr(self):
        switcher = {
            CompensationType.DieAlign: "DieAlign",
            CompensationType.Topography: "Topography",
            CompensationType.MapScan: "MapScan",
            CompensationType.AlignDie: "AlignDie",
            CompensationType.SkateDetection: "SkateDetection",
            CompensationType.OnTheFly: "OnTheFly",
            CompensationType.OffAxis: "OffAxis",
        }
        return switcher.get(self, "Invalid CompensationType")


class DefaultPattern(Enum):
    """A list of slots for visual patterns used by SENTIO.

    Each pattern is used for a specific purpose internally.

    Attributes:
        Align (0): The pattern used for wafer alignment.
        Home (1): Pattern used for finding the home position
        DieAlignPos1 (3): First pattern used for the die alignment. (on diced and taped wafers)
        DieAlignPos2 (4): Second pattern used for the die alignment. (on diced and taped wafers)
        TwoPoint (5): Pattern for Two-Point alignment
        Vce (6): Pattern used for VCE contact height detection.
        Ptpa (7): A pattern used for probe to pad alignment.
    """

    Align = 0
    Home = 1
    DieAlignPos1 = 3
    DieAlignPos2 = 4
    TwoPoint = 5
    Vce = 6
    Ptpa = 7

    def toSentioAbbr(self):
        switcher = {
            DefaultPattern.Align: "align",
            DefaultPattern.Home: "home",
            DefaultPattern.DieAlignPos1: "diealignpos1",
            DefaultPattern.DieAlignPos2: "diealignpos2",
            DefaultPattern.TwoPoint: "2pt",
            DefaultPattern.Vce: "vce",
            DefaultPattern.Ptpa: "ptpa",
        }
        return switcher.get(self, "Invalid default pattern id")


class DetectionAlgorithm(Enum):
    """Specifes one of several of SENTIO's internal detection algorithms.

    Not all versions of SENTIO support all algorithms. The most
    current versions of SENTIO have eliminated all models except
    for the ones based on deep learning.

    Those new Standard models detect multiple class of objects with
    far greater reliability than the outdated models based on haar
    cascades.

    Attributes:
        Keypoint (0): A Keypoint / ORB detector.
        ProbeDetector (1): Deep learning based AI model trained on various types of probe tips.
        WaferDetector (2): Deep learning based AI model trained on wafer structures.
    """

    Keypoint = 0
    ProbeDetector = 1
    WaferDetector = 2

    def toSentioAbbr(self):
        switcher = {
            DetectionAlgorithm.Keypoint: "Keypoint",
            DetectionAlgorithm.ProbeDetector: "ProbeDetector",
            DetectionAlgorithm.WaferDetector: "WaferDetector",
        }
        return switcher.get(self, "Invalid ProbeTipDetector")


class DetectionCoordindates(Enum):
    """Specifies the coordinate system used for reporting box detections.

    Used by SENTIO's built in DetectionAlgorithm

    Attributes:
        Image (0): Use image coordinates. Results are returned as pixel coordinates.
        Fov (1): Coordinates are in micrometer relative to the center of the field of view.
        Roi (2): Coordinates are in micrometer relative to the center of the region of interest.
    """

    Image = 0
    Fov = 1
    Roi = 2

    def toSentioAbbr(self):
        switcher = {
            DetectionCoordindates.Image: "Image",
            DetectionCoordindates.Fov: "Fov",
            DetectionCoordindates.Roi: "Roi",
        }
        return switcher.get(self, "Invalid DetectionCoordindates")


class DialogButtons(Enum):
    """A list of buttons that can be used in SENTIO's dialogs.

    Attributes:
        Ok (1): An Ok button.
        Yes (2): A Yes button.
        No (3): A No button.
        Cancel (4): A Cancel button.
        OkCancel (5): Both the Ok and the Cancel button.
        YesNo (6): Both a Yes and a No button.
        YesCancel (7): Yes and cancel button.
        YesNoCancel (8): Yes, No and Cancel button.
    """

    Ok = 1
    Yes = 2
    No = 3
    Cancel = 4
    OkCancel = 5
    YesNo = 6
    YesCancel = 7
    YesNoCancel = 8

    def toSentioAbbr(self):
        switcher = {
            DialogButtons.Ok: "Ok",
            DialogButtons.Cancel: "Cancel",
            DialogButtons.OkCancel: "OkCancel",
            DialogButtons.Yes: "Yes",
            DialogButtons.No: "No",
            DialogButtons.YesNo: "YesNo",
            DialogButtons.YesCancel: "YesCancel",
            DialogButtons.YesNoCancel: "YesNoCancel",
        }
        return switcher.get(self, "Invalid button id")


@deprecated("Use CompensationMode instead")
class DieCompensationMode(Enum):
    """Represents a compensation mode used by SENTIO.

    The compensation mode is a selector that defines what principal type of compensation shall be used.

    Attributes:
        Lateral (0): Use lateral compensation.
        Vertical (1): Use vertical compensation.
        Both (2): Use both lateral and vertical compensation.
        ProbeCard (3): Use probe card compensation.
        SkateDetection (4): Use skate detection
    """

    Lateral = 0
    Vertical = 1
    Both = 2
    ProbeCard = 3
    SkateDetection = 4

    def toSentioAbbr(self):
        switcher = {
            DieCompensationMode.Lateral: "Lateral",
            DieCompensationMode.Vertical: "Vertical",
            DieCompensationMode.Both: "Both",
            DieCompensationMode.ProbeCard: "ProbeCard",
            DieCompensationMode.SkateDetection: "SkateDetection",
        }
        return switcher.get(self, "Invalid DieCompensationMode function")


class DieCompensationType(Enum):
    """Compensation Type.

    Attributes:
        DieAlign (0): Die Alignment.
        MapScan (1): MapScan scans the wafer once and created a x,y compensation table for later use.
        Topography (2): Topography scans the height og the wafer on chuck and created a height map.
    """

    DieAlign = 0
    MapScan = 1
    Topography = 2

    def toSentioAbbr(self):
        switcher = {
            DieCompensationType.DieAlign: "DieAlign",
            DieCompensationType.MapScan: "MapScan",
            DieCompensationType.Topography: "Topography",
        }
        return switcher.get(self, "Invalid Compensation_Type function")


class DieNumber(Enum):
    """Specifies how dies are numbered.

    Attributes:
        Present (1): Number only dies that are present on the wafer.
        Selected (2): Number only dies that are selected for test.
    """

    Present = 1
    Selected = 2


@deprecated("ExecuteAction is deprecated.")
class ExecuteAction(Enum):
    Execute = 0
    Abort = 1

    def toSentioAbbr(self):
        switcher = {
            ExecuteAction.Execute: "execute",
            ExecuteAction.Abort: "abort",
        }
        return switcher.get(self, "Invalid ExecuteAction function")


@deprecated("ExecuteCompensation is deprecated.")
class ExecuteCompensation(Enum):
    AlignDie = 0
    MapScan = 1
    Topography = 2

    def toSentioAbbr(self):
        switcher = {
            ExecuteCompensation.AlignDie: "AlignDie",
            ExecuteCompensation.MapScan: "MapScan",
            ExecuteCompensation.Topography: "Topography",
        }
        return switcher.get(self, "Invalid compensation type")


class FindPatternReference(Enum):
    """Reference point for coordinates than returning a pattern position.

    Attributes:
        DieHome (0): Use Die Home position as reference.
        CenterOfRoi (1): Use Center of ROI as reference.
    """

    DieHome = 0
    CenterOfRoi = 1

    def toSentioAbbr(self):
        switcher = {
            FindPatternReference.DieHome: "DieHome",
            FindPatternReference.CenterOfRoi: "CenterOfRoi",
        }
        return switcher.get(self, "Invalid find pattern reference id")


class ImagePattern(Enum):
    align = 0
    home = 1
    diealignpos1 = 2
    diealignpos2 = 3
    twoPt = 4
    calc = 5

    def toSentioArg(self):
        switcher = {
            ImagePattern.align: "align",
            ImagePattern.home: "home",
            ImagePattern.diealignpos1: "diealignpos1",
            ImagePattern.diealignpos2: "diealignpos2",
            ImagePattern.twoPt: "2pt",
            ImagePattern.calc: "calc",
        }
        return switcher.get(self, "Invalid image pattern parameter")


class IMagProZReference(Enum):
    Zero = 0
    Relative = 1
    Center = 2

    def toSentioAbbr(self):
        switcher = {
            IMagProZReference.Zero: "Zero",
            IMagProZReference.Relative: "Relative",
            IMagProZReference.Center: "Center",
        }
        return switcher.get(self, "Invalid image pro z reference")


class LoaderStation(Enum):
    """An enumeration containing loader stations.

    Attributes:
        Cassette1 (0): First cassette station.
        Cassette2 (1): Second cassette station.
        PreAligner (2): Prealigner Station for wafer prealignment.
        Chuck (3): The chuck station.
        ForkA (4): Robot fork A
        ForkB (5): Roboit fork B
        WaferWallet (6): 5 slow wafer wallet station
        IdReader (7): Id-reader station.
    """

    Cassette1 = 0
    Cassette2 = 1
    PreAligner = 2
    Chuck = 3
    ForkA = 4
    ForkB = 5
    WaferWallet = 6
    IdReader = 7

    def toSentioAbbr(self):
        switcher = {
            LoaderStation.Cassette1: "cas1",
            LoaderStation.Cassette2: "cas2",
            LoaderStation.PreAligner: "pa",
            LoaderStation.Chuck: "chuck",
            LoaderStation.ForkA: "forka",
            LoaderStation.ForkB: "forkb",
            LoaderStation.WaferWallet: "ww",
        }
        return switcher.get(self, "Invalid loader station id")


class LoadPosition(Enum):
    """An enumeration containing the possible load positions.

    Not all load positions are available on all probe stations.

    Attributes:
        Front (0): The Front Load position. All probe station have a front load position.
        Side (1): The Side Load position. The side load position is optional and only present on systems with a side loader, a cassette loader or a wafer wallet.
    """

    Front = 0
    Side = 1

    def toSentioAbbr(self):
        switcher = {LoadPosition.Front: "front", LoadPosition.Side: "side"}
        return switcher.get(self, "Invalid Load position")


class Module(Enum):
    """An enumerator containing the names of all SENTIO modules.

    Module availability is determined by the specific type of probe station.
    All probe stations have the dashboard, setup and service modules. Most likely
    wafermap and Vision module will also be present (unless you have a purely manual
    station)

    Attributes:
        Wafermap (0): The wafermap module. This module is used for wafer alignment and wafer mapping.
        Vision (1): The vision module. This module is used for die alignment and wafer inspection.
        Setup (2): The setup module. This module is used for setting up the probe station.
        Service (3): The service module. This module is used for service and maintenance.
        Qalibria (4): The Qalibria module. This module is used for Qalibria integration.
        AuxSites (5): The aux site module. This module is used for auxilliary site integration.
        Loader (6): The loader module. This module is used for loader integration.
        Dashboard (7): The dashboard module. This module is used for dashboard integration.
    """

    Wafermap = 0
    Vision = 1
    Setup = 2
    Service = 3
    Qalibria = 4
    AuxSites = 5
    Loader = 6
    Dashboard = 7

    def toSentioAbbr(self):
        switcher = {
            Module.Wafermap: "Wafermap",
            Module.Vision: "Vision",
            Module.Setup: "Setup",
            Module.Service: "Service",
            Module.Qalibria: "Qalibria",
            Module.AuxSites: "AuxSites",
            Module.Loader: "Loader",
            Module.Dashboard: "Dashboard",
        }
        return switcher.get(self, "Invalid Module Name")


@deprecated(reason="duplicated; Use CompensationMode instead.")
class OnTheFlyMode(Enum):
    Lateral = 0
    Vertical = 1
    Both = 2
    ProbeCard = 3

    def toSentioAbbr(self):
        switcher = {
            OnTheFlyMode.Lateral: "AlignDie",
            OnTheFlyMode.Vertical: "MapScan",
            OnTheFlyMode.Both: "Topography",
            OnTheFlyMode.ProbeCard: "ProbeCard",
        }
        return switcher.get(self, "Invalid OTF mode")


class OrientationMarker(Enum):
    """Defines the wafers orientation marker.

    Orientation markers can either be a notch or a flat in the wafer.
    The orientation marker is used by the prealigner to detect
    the wafer position.

    Today mostly notches are used ar orientation markes as they reduce the
    waste of wafer space.

    Attributes:
        Notch (0): Wafer uses a notch.
        Flat (1): Wafer uses a flat.
    """

    Notch = 0
    Flat = 1

    def toSentioAbbr(self):
        switcher = {OrientationMarker.Notch: "Notch", OrientationMarker.Flat: "Flat"}
        return switcher.get(self, "Invalid orientation marker")


class PoiReferenceXy(Enum):
    """Referenc position for points of interest.

    Each point of interest can either be defines with respect to the center of the stage.
    Or with respect to the center of the die.

    Attributes:
        DieCenter (0): Use die center as the position reference.
        StageCenter (1): Use stage center as the position reference.
    """

    DieCenter = 0
    StageCenter = 1

    def toSentioAbbr(self):
        switcher = {
            PoiReferenceXy.DieCenter: "DieCenter",
            PoiReferenceXy.StageCenter: "StageCenter",
        }
        return switcher.get(self, "Invalid stage")


class ProjectFileInfo(Enum):
    """An enumerator containing the different aspects of retrieving current project info.

    Attributes:
        NameOnly (0): Return only the project name.
        FullPath (1): Return the full path to the project file.
    """

    NameOnly = 0
    FullPath = 1

    def toSentioAbbr(self):
        switcher = {
            ProjectFileInfo.NameOnly: "Name",
            ProjectFileInfo.FullPath: "FullPath",
        }
        return switcher.get(self, "Invalid ProjectFileInfo")


class PtpaType(Enum):
    """Defines the type of Probe to Pad Alignment used by SENTIO.

    Attributes:
        OffAxis (0): Use off-axis PTPA with the platen camera and the chuck camera looking up to the probe tips.
        OnAxis (1): Use on axis PTPA with the scope camera looking down on the probe tips.
    """

    OffAxis = 0
    OnAxis = 1

    def toSentioAbbr(self):
        switcher = {PtpaType.OffAxis: "offaxis", PtpaType.OnAxis: "onaxis"}
        return switcher.get(self, "Invalid ptpa type")


class SnapshotType(Enum):
    """Specifies the type of image snapshot to be taken"""

    CameraRaw = 0
    """ The snapshot is the raw image from the camera in original resolution """

    WithOverlays = 1
    """ The snapshot consists of a screenshot of the vision module inside SENTIO. This method 
        will also contain the overlays displayed by the vision module but the image resolution
        will be whatever the current resolution of SENTIO's vision module is. """

    def toSentioAbbr(self):
        """Turn the SnapshotType into a string that can be used as a parameter for SENTIO's snap_image command."""
        switcher = {SnapshotType.CameraRaw: "0", SnapshotType.WithOverlays: "1"}
        return switcher.get(self, "Invalid SnapshotType type")


class SnapshotLocation(Enum):
    """Specifies where a snapshot shall be saved."""

    Prober = 0
    """ The snapshot is stored remotely on the prober PC that is executing SENTIO. """

    Local = 1
    """ The snapshot will be transferred to the PC that is running the python control script. 
        This option only makes sense if the prober and the control PC are different. """


class Stage(Enum):
    """Represents a stage in SENTIO.

    A stage is a piece of hardware that can be controlled by motors and moved in x, y and
    probably also z direction.

    Attributes:
        Chuck (0): The chuck stage. This is where the wafer is placed.
        Scope (1): The scope stage controls the downward looking microscope. This is an optional stage but it will be present on most probe stations.
        Vce (2): First Vce stage. Vce Stages can only be moved in z-direction.
        Vce2 (3): Second Vce stage. Vce Stages can only be moved in z-direction.
        Probe1 (4): First motorized probe.
        Probe2 (5): Second motorized probe.
        Probe3 (6): Third motorized probe.
        Probe4 (7): Fourth motorized probe.
    """

    Chuck = 0
    Scope = 1
    Vce = 2
    Vce2 = 3
    Probe1 = 4
    Probe2 = 5
    Probe3 = 6
    Probe4 = 7

    def toSentioAbbr(self):
        switcher = {
            Stage.Chuck: "chuck",
            Stage.Scope: "scope",
            Stage.Vce: "vce01",
            Stage.Vce2: "vce02",
            Stage.Probe1: "Probe01",
            Stage.Probe2: "Probe02",
            Stage.Probe3: "Probe03",
            Stage.Probe4: "Probe04",
        }
        return switcher.get(self, "Invalid stage")


class SteppingContactMode(Enum):
    """This mode defines how the chuck behaves during steeping.

    Attributes:
        BackToContact (0): Chuck will move back to contact position after stepping.
        StepToSeparation (1): Chuck will move to separation position after stepping.
        LockContact (2): Chuck cannot step when at contact. You will have to manually move it away from its contact position before issuing the next step command.
    """

    BackToContact = 0
    StepToSeparation = 1
    LockContact = 2

    def toSentioAbbr(self):
        switcher = {
            SteppingContactMode.BackToContact: "BackToContact",
            SteppingContactMode.StepToSeparation: "StepToSeparation",
            SteppingContactMode.LockContact: "LockContact",
        }
        return switcher.get(self, "Invalid stepping mode")


class TestSelection(Enum):
    """Specifies which dies shall be selected for test.

    If a die is selected for test it is activated in the wafer map.

    Attributes:
        Nothing (0): Select no dies.
        Good (1): Select only the good dies.
        GoodAndUgly (2): Select only the good dies. Ugly dies are the dies with at least one corner in the edge area of the map. The are completely present on the wafer but may be damaged or have incomplete structures.
        GoodUglyAndEdge (3): Good dies are the dies that are completely within the wafer map with no edge lying in the edge area of the map. Ugly dies are the dies with at least one corner in the edge area of the map. The are completely present on the wafer but may be damaged or have incomplete structures. The dies with at leas one edge outside of the wafer are called edge dies. Those dies are incomplete.
        All (4): Select all dies for testing.
    """

    Nothing = 0
    Good = 1
    GoodAndUgly = 2
    GoodUglyAndEdge = 3
    All = 4

    def toSentioAbbr(self):
        switcher = {
            TestSelection.Nothing: "n",
            TestSelection.Good: "g",
            TestSelection.GoodAndUgly: "u",
            TestSelection.GoodUglyAndEdge: "e",
            TestSelection.All: "a",
        }
        return switcher.get(self, "Invalid TestSelection")


class ScopeXYReference(Enum):
    """Position reference for scope motions.

    Attributes:
        Zero (0): Use absolute scope coordinates.
        Home (1): Use coordinates with respect to the scope home position.
        Relative (2): Use coordinates with respect to the current scope position.
    """

    Zero = 0
    Home = 1
    Relative = 2

    def toSentioAbbr(self):
        switcher = {
            ScopeXYReference.Zero: "Z",
            ScopeXYReference.Home: "H",
            ScopeXYReference.Relative: "R",
        }
        return switcher.get(self, "Invalid scope xy reference")


class ScopeZReference(Enum):
    """Scope z reference for scoep motions.

    Attributes:
        Zero (0): Use absolute scope coordinates.
        Relative (1): Use coordinates with respect to the current scope position.
    """

    Zero = 0
    Relative = 1

    def toSentioAbbr(self):
        switcher = {ScopeZReference.Zero: "Z", ScopeZReference.Relative: "R"}
        return switcher.get(self, "Invalid scope z reference")


class ProbeSentio(Enum):
    """An enumeration containing a list of position for motorized probes.

    Attributes:
        East (0): Probe is on the right side of the chuck.
        West (1): Probe is on the left side of the chuck.
        North (2): Probe is at the back side of the prober.
        South (3): Probe is on the front side of the prober.
    """

    East = 0
    West = 1
    North = 2
    South = 3

    def toSentioAbbr(self):
        switcher = {
            ProbeSentio.East: "East",
            ProbeSentio.West: "West",
            ProbeSentio.North: "North",
            ProbeSentio.South: "South",
        }
        return switcher.get(self, "Invalid ProbeSentio enumerator")


class ProbeXYReference(Enum):
    """Position reference for mororized probe movements.

    Attributes:
        Zero (0): Use absolute probe coordinates.
        Home (1): Use coordinates with respect to the probe home position.
        Current (2): Use coordinates with respect to the current probe position.
    """

    Zero = 0
    Home = 1
    Current = 2

    def toSentioAbbr(self):
        switcher = {
            ProbeXYReference.Zero: "Zero",
            ProbeXYReference.Home: "Home",
            ProbeXYReference.Current: "Current",
        }
        return switcher.get(self, "Invalid probe xy reference")


class ProbeZReference(Enum):
    """Position reference for probe z motions.

    Attributes:
        Zero (0): Use absolute probe coordinates.
        Current (1): Use coordinates with respect to the current probe position.
        Contact (2): Use coordinates with respect to the contact height.
        Separation (3): Use coordinates with respect to the separation height.
    """

    Zero = 0
    Current = 1
    Contact = 2
    Separation = 3

    def toSentioAbbr(self):
        switcher = {
            ProbeZReference.Zero: "Zero",
            ProbeZReference.Current: "Current",
            ProbeZReference.Contact: "Contact",
            ProbeZReference.Separation: "Separation",
        }
        return switcher.get(self, "Invalid probe z reference")


class PtpaFindTipsMode(Enum):
    """Specifies the mode used by ptpa tip finding.

    Attributes:
        OnAxis (0): Use on axis tip finding.
        OffAxis (1): Use off axis tip finding.
    """

    OnAxis = 0
    OffAxis = 1

    def toSentioAbbr(self):
        switcher = {
            PtpaFindTipsMode.OnAxis: "OnAxis",
            PtpaFindTipsMode.OffAxis: "OffAxis",
        }
        return switcher.get(self, "Invalid PTPA_Find_Tips_Mode function")


class RemoteCommandError:
    """A list of possible error codes used by SENTIO.

    This list may not contain the full list of codes used by SENTIO due to
    the fact that SENTIO is a moving target and new error codes may be added
    or removed.

    Have a look at the remote command specification to see which codes are
    supported by your version.

    Attributes:
        NoError (0): No error occured.
        InternalError (1): An internal error occured in SENTIO. This is not supposed to happen and you can probably not fix the issue on your own. Please contact SENTIO support.
        ExecutionError (2): A generic execution error. This is the most widely used error code to signal remote command failure.
        CommandHandlerNotFound (3): A command handler for a certain subsystem of SENTIO was not found. This may happen when sending commands to a SENTIO module that is not available on a given machine.
        InvalidCommand (4): not used by SENTIO's native remote command set.
        InvalidCommandFormat (5): not used by SENTIO's native remote command set.
        InvalidParameter (6): A remote command parameter is incoreect.
        InvalidNumberOfParameters (7): The number of submitted remote command parameters is incorect.
        ArgumentOutOfBounds (8): A submitted parameter exceeds the range of allowed values.
        FileNotFound (9): A file that was supposed to be loaded by SENTIO was not found.
        InvalidFileFormat (10): not used by SENTIO's native remote command set.
        EndOfRoute (11): Stepping reached the end of the route.
        InvalidOperation (12): The requested operation is not allowed in the current state.
        NotSupported (13): The requested operation is not supported by the current version of SENTIO.
        SubsiteNotRoutable (14): The requested subsite is not routable.
        ProjectRequired (15): A requested operation require an active project.
        Unused (16): This error code is unused.
        PrealignmentFailed (17): Prealignment failed.
        HomePositionNotSet (18): The home position is not set.
        Timeout (19): A command or operation timed out.
        PatternNotTrained (20): A required pattern is not trained
        PatternNotFound (21): A pattern could not be found
        TooManyPatternsFound (22): Too many patterns were found
        ContactHeightNotSet (23): The contact height is not set.
        AutoFocusFailed (24): Auto focus failed on the wafer.
        TipFocusFailed (25): Auto focus failed on the tips.
        TipNotFound (26): A tip could not be found.
        OffsetOverTolerance (27): The offset is over tolerance.
        CommandPending (30): Returned when the status of an async command is polled with query_command_status and the command is Running
        AsyncCommandAborted (31): Returned when a async command was aborted prematurely
        UnknownCommandId (32): Returned when an async command is queried but SENTIO does not know anything about this command id
        CameraNotCalibrated (35): A camera required for a vision task is not calibrated
        CameraDoesNotExist (36): A required camera is not installed in the system.
        AlignAccuracyBad (37): Alignment accuracy over 10 µm
        SteppingWithWrongZPosition (38): Stepping when Chuck is below Separation
        FrontDoorOpen (60): The front load door is open
        LoaderDoorOpen (61): The side door is open
        FrontDoorLockFail (62): Front door lock cannot be engaged
        LoaderDoorLockFail (63): Side door lock cannot be engaged
        SlotOrStationOccupied (64): A slot or station that is the target of a wafer transfer is already occupied
        SlotOrStationEmpty (65): A slot or station that is the origin of a wafer transfer does not have a wafer
        ProbeBackDoorOpen (66): The probe back door is open
        ProbeSideDoorOpen (67): The probe side door is open (Not loader door, only TS2500/SE has this door, the door near the probe back door)
        VacuumFailed (68): Vacuum failed
        LoaderTrayDoorOpen (69): The tray door is open
        LoaderCassetteDoesNotExist (80): The cassette does not exist
        LoaderSlotNumberError (81): The slot number is not correct
        LoaderPreAlignerAngleError (83): The prealigner angle is not correct
    """

    NoError = 0
    InternalError = 1
    ExecutionError = 2
    CommandHandlerNotFound = 3
    InvalidCommand = 4
    InvalidCommandFormat = 5
    InvalidParameter = 6
    InvalidNumberOfParameters = 7
    ArgumentOutOfBounds = 8
    FileNotFound = 9
    InvalidFileFormat = 10
    EndOfRoute = 11
    InvalidOperation = 12
    NotSupported = 13
    SubsiteNotRoutable = 14
    ProjectRequired = 15
    EndOfList = 16
    PrealignmentFailed = 17
    HomePositionNotSet = 18
    Timeout = 19
    PatternNotTrained = 20
    PatternNotFound = 21
    TooManyPatternsFound = 22
    ContactHeightNotSet = 23
    AutoFocusFailed = 24
    TipFocusFailed = 25
    TipNotFound = 26
    OffsetOverTolerance = 27
    CommandPending = 30
    AsyncCommandAborted = 31
    UnknownCommandId = 32
    CameraNotCalibrated = 35
    CameraDoesNotExist = 36
    AlignAccuracyBad = 37
    SteppingWithWrongZPosition = 38
    FrontDoorOpen = 60
    LoaderDoorOpen = 61
    FrontDoorLockFail = 62
    LoaderDoorLockFail = 63
    SlotOrStationOccupied = 64
    SlotOrStationEmpty = 65
    ProbeBackDoorOpen = 66
    ProbeSideDoorOpen = 67
    VacuumFailed = 68
    LoaderTrayDoorOpen = 69
    LoaderCassetteDoesNotExist = 80
    LoaderSlotNumberError = 81
    LoaderPreAlignerAngleError = 83
    LoaderNoWaferOnPrealigner = 85
    LoaderNoWaferOnChuck = 86
    LoaderNoWaferAtSlotOrTray = 87
    LoaderNoWaferOnRobot = 88
    LoaderNoIdReader = 90
    LoaderReadIdFail = 91
    LoaderTransferWaferFail = 92
    LoaderNoWaferOnAuxiliary = 93

    ProbeNotInitialized = 100
    ProbeServoOnOffFail = 101

    OvertravelOutOfAxisLimit = 120
    MissingTopographyTable = 121
    IndexerDeviceIsNotExisting = 122

    LoaderWaferOnFork = 150
    LoaderWaferSlideOutCassette = 151
    LoaderNoWaferOnFork = 152
    LoaderWaferTrackerReceivedTimeout = 153
    LoaderWaferIsOblique = 154
    LoaderCassetteWithoutScan = 155
    LoaderPrealignerSensorSwitchSizeTimeOut = 156
    LoaderCassetteDoorBarOpenTimeout = 157
    LoaderCassetteDoorBarCloseTimeout = 158
    LoaderWaferIsCrossOver = 159
    LoaderWaferOnChuck = 160
    LoaderPrealignCalOffsetTooLarge = 161
    LoaderPrealignFailedTooMuchTime = 162
    LoaderConnectPrealignerFailed = 163
    LoaderWaferOnTray = 164
    LoaderCassetteMappingLagCountTooMuch = 165
    LoaderTransferSoakingTimeOut = 166
    LoaderUpdateMappingResultFail = 167
    LoaderTrayDoorLockTimeout = 168
    LoaderWaferOnPrealigner = 169

    ThermalSoakingTimeIsNotCorrect = 200

    SiPhMoveHoverFail = 300
    SiPhMoveSeparationFail = 301
    SiPhGradientSearchFail = 302
    SiPhFastAlignFail = 303
    SiPhPowerMeasurementFail = 304
    SiPhCouplingFail = 305
    SiPhTrackingFail = 306
    SiPhPivotPointSearchFail = 307

    PositionerSiteDataFail = 400

    QaChuckNotWorkingPosition = 450
    QaSubstrateNotSet = 451
    QaRemoteModeNotSet = 452
    QaStandardsEmpty = 453
    QaCalculateEtsFail = 454
    QaSetCalDriftDutDataFail = 455
    QaCalDriftDutNotFound = 456
    QaCalDriftDutFail = 457
    QaSwitchViewFail = 458
    QaExportCalDriftDataFail = 459


class RoutingPriority(Enum):
    """Defines the stepping order.

    Attributes:
        RowUniDir (0): Rows first always step in same direction.
        ColUniDir (1): Columns first always step in same direction.
        RowBiDir (2): Rows first. Step odd rows backwards.
        ColBiDir (3): Columns first. Step odd columns backwards.
    """

    RowUniDir = 0
    ColUniDir = 1
    RowBiDir = 2
    ColBiDir = 3

    def toSentioAbbr(self):
        switcher = {
            RoutingPriority.RowUniDir: "r",
            RoutingPriority.ColUniDir: "c",
            RoutingPriority.RowBiDir: "wr",
            RoutingPriority.ColBiDir: "wc",
        }
        return switcher.get(self, "Invalid RoutingPriority enumerator")


class RoutingStartPoint(Enum):
    """Defines the starting point for routing (stepping commands).

    Attributes:
        UpperLeft (0): Stepping starts in upper left corner of the map.
        UpperRight (1): Stepping starts in upper right corner of the map.
        LowerLeft (2): Stepping stars in lower left corner of the map.
        LowerRight (3): Stepping starts in lower right corner of the map.
    """

    UpperLeft = 0
    UpperRight = 1
    LowerLeft =  2
    LowerRight = 3

    def toSentioAbbr(self):
        switcher = {
            RoutingStartPoint.UpperLeft: "ul",
            RoutingStartPoint.UpperRight: "ur",
            RoutingStartPoint.LowerLeft: "ll",
            RoutingStartPoint.LowerRight: "lr",
        }
        return switcher.get(self, "Invalid RoutingStartPoint enumerator")


class StatusBits:
    """List of status codes used by SENTIO.

    SENTIO will encode certain status information into the error code variable.
    This status information does not represent an errors but rather a machine
    status.

    Attributes:
        EndOfRoute (1): Stepping reached the end of the route.
        LastSite (2): Stepping reached the last site of a die.
    """

    EndOfRoute = 1
    LastSite = 2


class SoftwareFence(Enum):
    """An enumerator holding software fence implementations.

    Attributes:
        Disabled (0): Fence is disabled.
        Round (1): A round fence around the chuck, excluding aux sites.
        Rectangle (2): A rectangular fence around the chuck, may include parts of the aux sites.
        SoftwareLimit (3): Use software limits on axis. A Large rectangular fence around the chuck motion ares. Collisions with the prober housing are possible (TS-2000; tilted front door)
    """

    Disabled = 0
    Round = 1
    Rectangle = 2
    SoftwareLimit = 3

    def toSentioArg(self):
        switcher = {
            SoftwareFence.Disabled: "Disable",
            SoftwareFence.Rectangle: "Rectangle",
            SoftwareFence.Round: "Round",
            SoftwareFence.SoftwareLimit: "SoftwareLimit",
        }
        return switcher.get(self, "Invalid SoftwareFence parameter")


class ThermoChuckState(Enum):
    """The state of a thermo chuck.

    Attributes:
        Soaking (0): The chuck is in soaking state,
        Cooling (1): The chuck is in cooling state,
        Heating (2): The chuck is in heating state,
        Uncontrolled (3): The chuck is in uncontrolled state,
        Standby (4): The chuck is in standby state,
        Error (5): The chuck is in error state,
        Controlling (6): The chuck is in controlling state,
    """

    Soaking = 0
    Cooling = 1
    Heating = 2
    Uncontrolled = 3
    Standby = 4
    Error = 5
    Controlling = 6
    Unknown = 7


class VceZReference(Enum):
    """Reference for Vce z motions.

    Attributes:
        Zero (0): Use absolute Vce coordinates.
        Relative (1): Use coordinates with respect to the current Vce position.
    """

    Zero = 0
    Relative = 1

    def toSentioAbbr(self):
        switcher = {VceZReference.Zero: "Z", VceZReference.Relative: "R"}
        return switcher.get(self, "Invalid vce z reference")


class VirtualCarrierInitFlags(Enum):
    """Flags for initializing a virtual carrier measurement.

    Attributes:
        Start (0): Start a new virtual carrier measurement.
        Continue (1): Continue an existing virtual carrier measurement. If no measurement is running, this will start a new one.
    """

    Start = 0
    Continue = 1

    def toSentioAbbr(self):
        switcher = {
            VirtualCarrierInitFlags.Start: "Start",
            VirtualCarrierInitFlags.Continue: "Continue",
        }
        return switcher.get(self, "Invalid VirtualCarrierInitFlags")


class VirtualCarrierStepProcessingState(Enum):
    """State of a single virtual carrier processing step. 
    
    Attributes:
        Skip (0): The step is skipped.
        Done (1): The step is done.
        Ready (2): The step is ready to be executed.
    """
    Skip = 0
    Done = 1
    Ready = 2

    def toSentioAbbr(self):
        switcher = {
            VirtualCarrierStepProcessingState.Skip: "Skip",
            VirtualCarrierStepProcessingState.Done: "Done",
            VirtualCarrierStepProcessingState.Ready: "Ready",
        }
        return switcher.get(self, "Invalid VirtualCarrierStepProcessingState")


class WorkArea(Enum):
    """An enumeration containing probe station work areas. The OffAxis work area is only present if the specific model of probe station supports it.

    Attributes:
        Probing (0): The probing work area is the area in which the chuck is under the downward looking microscope. This is where the wafer is probed.
        Offaxis (1): The off axis work area is the area in which the chuck is under the off axis camera. This is where off axis ptpa is performed. The wafer cannot be probed here because there is no probe card.
    """

    Probing = 0
    Offaxis = 1

    def toSentioAbbr(self):
        switcher = {
            WorkArea.Probing: "Probing",
            WorkArea.Offaxis: "Offaxis",
        }
        return switcher.get(self, "Invalid chuck site")


class ZPositionHint(Enum):
    """Represents a hint for the z position of a stage.

    Not all values are used by all stages. A scope does not have a contact height
    and a chuck Hover height may be disabled by SENTIO.

    Attributes:
        Default (0): Used internally only. Essentially means the value is unset or undefined.
        Contact (1): Stage is at contact position.
        Hover (2): Stage is at hover position.
        Separation (3): Stage is at separation position
        Lift (4): Stage is at Lift position.
        Transfer (5): Chuck is at transfer position. This is used for the chuck only when the loader is doing a wafer transfer internally.
    """

    Default = 0
    Contact = 1
    Hover = 2
    Separation = 3
    Lift = 4
    Transfer = 5

    def toSentioAbbr(self):
        switcher = {
            ZPositionHint.Default: "Default",
            ZPositionHint.Contact: "Contact",
            ZPositionHint.Hover: "Hover",
            ZPositionHint.Separation: "Separation",
            ZPositionHint.Lift: "Lift",
            ZPositionHint.Transfer: "Transfer",
        }
        return switcher.get(self, "Invalid ZPositionHint")



from abc import ABC, abstractmethod

from sentio_prober_control.Communication.CommunicatorBase import CommunicatorBase


class ProberException(Exception):
    """A custom exception class for SENTIO's python package.

    All errors that occur in this package will be wrapped in this exception class.
    """

    def __init__(self, msg: str, errc: int = -1):
        """Create a new ProberException object.

        Args:
            msg (str): The error message.
            errc (int): The error code (optional). If no error code is set -1 is used.
        """
        super().__init__(msg)
        self.__message = msg
        self.__error = errc

    def error(self):
        """Get error code.

        Returns:
            errc (int): The error code of the exception.
        """
        return self.__error

    def message(self):
        """Returns the error message.

        Returns:
            errorcode (str): The error message.
        """
        return self.__message


class ProberBase(ABC):
    """Abstract base class of a probe station."""

    def __init__(self, comm: CommunicatorBase):
        """Creates a new ProberBase object.

        Attributes:
            comm (CommunicatorBase): The communication object used to communicate with the probe station.

        Args:
            comm (CommunicatorBase): The communication object used to communicate with the probe station.
        """

        self.__comm = comm


    @property
    def comm(self) -> CommunicatorBase:
        """Get the communicator object.

        Returns:
            comm (CommunicatorBase): The communicator object.
        """
        return self.__comm
    

    @abstractmethod
    def name(self) -> str:
        """Returns the name of the probe station.

        Must be implemented by derived classes.

        Returns:
            name (str): The name of the probe station.
        """
        pass



from typing import Tuple

from sentio_prober_control.Sentio.Enumerations import ProbeSentio
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase


class SiPHCommandGroup(CommandGroupBase):
    """This command group contains functions for working with SiPH applications.
    You are not meant to instantiate this class directly. Access it via the siph attribute
    of the [SentioProber](SentioProber.md) class.
    """

    def __init__(self, comm) -> None:
        super().__init__(comm)


    def fast_alignment(self) -> Response:
        """Perform fast fiber alignment."""

        self.comm.send("siph:fast_alignment")
        return Response.check_resp(self.comm.read_line())


    def get_cap_sensor(self) -> Tuple[float, float]:
        """Get the capacitance sensor value.

        :raises: ProberException if an error occured.
        :return: A tuple with the values from the capacity sensors of probe 1 and probe 2.
        """
        self.comm.send("siph:get_cap_sensor")
        resp = Response.check_resp(self.comm.read_line())

        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1])


    def get_intensity(self) -> float:
        """Get the current intensity value.

        :raises: ProberException if an error occured.
        :returns: The intensity value.
        """

        self.comm.send("siph:get_intensity")
        resp = Response.check_resp(self.comm.read_line())
        return int(resp.message())


    def gradient_search(self) -> None:
        """Execute SiPh gradient search.

        :raises: ProberException if an error occured.
        """

        self.comm.send("siph:gradient_search")
        Response.check_resp(self.comm.read_line())


    def move_hover(self, probe: ProbeSentio) -> None:
        """Move SiPh probe to hover height.

        :param probe: The probe on which the SiPh probe is mounted.
        :raises: ProberException if an error occured.
        """

        self.comm.send(f"siph:move_hover {probe.toSentioAbbr()}")
        Response.check_resp(self.comm.read_line())


    def move_separation(self, probe: ProbeSentio) -> None:
        """Move SiPh probe to separation height.

        :param probe: The probe on which the SiPh probe is mounted.
        :raises: ProberException if an error occured.
        """

        self.comm.send(f"siph:move_separation {probe.toSentioAbbr()}")
        Response.check_resp(self.comm.read_line())



from deprecated import deprecated
from typing import Tuple

from sentio_prober_control.Sentio.Enumerations import LoaderStation, OrientationMarker, RemoteCommandError
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase
from sentio_prober_control.Sentio.CommandGroups.LoaderVirtualCarrierCommandGroup import LoaderVirtualCarrierCommandGroup
from sentio_prober_control.Communication.CommunicatorBase import CommunicatorBase
from sentio_prober_control.Sentio.ProberBase import ProberException

class LoaderCommandGroup(CommandGroupBase):
    """This command group contains functions for working with the loader.
    You are not meant to instantiate this class directly. Access it via the loader attribute
    of the [SentioProber](SentioProber.md) class.

    Example:

    ```py
    from sentio_prober_control.Sentio.ProberSentio import SentioProber

    prober = SentioProber.create_prober("tcpip", "127.0.0.1:35555")
    scan_result = prober.loader.scan_station(LoaderStation.Cassette1)
    ```
    """

    def __init__(self, parent : 'SentioProber') -> None:
        super().__init__(parent)

        self.vc: LoaderVirtualCarrierCommandGroup = LoaderVirtualCarrierCommandGroup(self)


    def start_prepare_station(self, station: LoaderStation, angle: float | None = None) -> Response:
        """Prepare a loader station for wafer processing.

        This command will trigger a scan of all wafers in a station. This command is an async
        remote command!

        Wraps Sentios "loader:start_prepare_station" remote command.

        Args:
            station (LoaderStation): The station to prepare.
            angle (float): The prealignment angle. If None, no prealignment will be done.

        Returns:
            response (Response): A Response object.
        """

        if angle == None:
            self.comm.send(f"loader:start_prepare_station {station.toSentioAbbr()}")
        else:
            self.comm.send(f"loader:start_prepare_station {station.toSentioAbbr()}, {angle}")

        return Response.check_resp(self.comm.read_line())


    def scan_station(self, station: LoaderStation) -> str:
        """Scans a loader station for wafers.
        Wraps Sentios "loader:scan_station" remote command.

        Args:
            station (LoaderStation): The station to scan.

        Returns:
            result (str): A string with the scan result.
        """

        self.comm.send(f"loader:scan_station {station.toSentioAbbr()}")
        resp = Response.check_resp(self.comm.read_line())
        return resp.message()


    def has_station(self, station: LoaderStation) -> bool:
        """Check wether a certain loader station is present.

        Wraps Sentios "loader:has_station" remote command.

        Args:
            station (LoaderStation): The station to check.

        Returns:
            has_station (bool): True if the station is present, False otherwise.
        """

        self.comm.send(f"loader:has_station {station.toSentioAbbr()}")
        resp = Response.check_resp(self.comm.read_line())
        return resp.message() == "1"


    def transfer_wafer(
        self,
        src_station: LoaderStation,
        src_slot: int,
        dst_station: LoaderStation,
        dst_slot: int
    ):
        """Transfer a wafer from one loader station to another.

        Args:
            src_station (LoaderStation): The source station.
            src_slot (int): The source slot.
            dst_station (LoaderStation): The destination station.
            dst_slot (int): The destination slot.
        """

        self.comm.send(
            f"loader:transfer_wafer {src_station.toSentioAbbr()}, {src_slot}, {dst_station.toSentioAbbr()}, {dst_slot}"
        )
        return Response.check_resp(self.comm.read_line())


    def load_wafer(self, src_station: LoaderStation, src_slot: int, angle: int | None = None):
        """Load a wafer onto the chuck with optional prealignment.

        Args:
            src_station: The source station.
            src_slot: The source slot.
            angle: The prealignment angle.

        Returns:
            response (Response): A Response object.
        """

        if angle is None:
            self.comm.send(f"loader:load_wafer {src_station.toSentioAbbr()}, {src_slot}")
        else:
            self.comm.send(f"loader:load_wafer {src_station.toSentioAbbr()}, {src_slot}, {angle}")

        return Response.check_resp(self.comm.read_line())


    def prealign(self, marker: OrientationMarker, angle: int):
        """Prealign a wafer.

        Wraps Sentios "loader:prealign" remote command.

        Args:
            marker (OrientationMarker): The type of wafer orientation marker.
            angle (float): The prealignment angle.

        Returns:
            response (Response): A Response object.
        """

        self.comm.send(f"loader:prealign {marker.toSentioAbbr()}, {angle}")
        return Response.check_resp(self.comm.read_line())

    def query_wafer_status(self, station : LoaderStation, slot : int) -> Tuple[LoaderStation, int, int, int, float] | None:
        """Query the status of a wafer in a loader station.

        Wraps Sentios "loader:query_wafer_status" remote command.

        Args:
            station (LoaderStation): The station to query.
            slot (int): The slot to query.

        Returns:
            status (Tuple[LoaderStation, int, int, int, float]): A tuple with the following elements: OriginStation, OriginSlot, Wafer Size, Wafer Orientation, Progress Value.
        """

        self.comm.send(f"loader:query_wafer_status {station.toSentioAbbr()}, {slot}")
        
        try:
            resp = Response.check_resp(self.comm.read_line())
        except ProberException as e: 
            if e.error() == RemoteCommandError.SlotOrStationEmpty:
                return None
            else:
                raise
        
        tok = resp.message().split(',')
        origin_station = LoaderStation[tok[0]]
        origin_slot = int(tok[1])
        
        try:
           # SENTIO may return NaN for unknown size
           size = int(tok[2])
        except:
            size = -1

        orient = int(tok[3])
        progress = float(tok[4])

        return (origin_station, origin_slot, size, orient, progress)
    

    @deprecated("duplicate functionality; Use SentioProber.move_chuck_work_area!")
    def switch_work_area(self, area: str):
        self.comm.send("move_chuck_work_area {0}".format(area))
        resp = Response.check_resp(self.comm.read_line())
        return resp.message()


    def unload_wafer(self) -> None:
        """Unload the current wafer from the chuck.

        Wraps Sentios "loader:unload_wafer" remote command.
        """

        self.comm.send("loader:unload_wafer")
        Response.check_resp(self.comm.read_line())



import base64
import os
from typing import Tuple
from enum import Enum

from deprecated import deprecated

from sentio_prober_control.Sentio.Enumerations import (
    ChuckSite,
    ChuckXYReference,
    ChuckThetaReference,
    ChuckZReference,
    DialogButtons,
    LoadPosition,
    ProjectFileInfo,
    Module,
    ScopeXYReference,
    ScopeZReference,
    Stage,
    SteppingContactMode,
    VceZReference,
    WorkArea
)

from sentio_prober_control.Sentio.ProberBase import ProberBase, ProberException
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Communication.CommunicatorBase import CommunicatorBase
from sentio_prober_control.Communication.CommunicatorGpib import CommunicatorGpib, GpibCardVendor
from sentio_prober_control.Communication.CommunicatorTcpIp import CommunicatorTcpIp
from sentio_prober_control.Communication.CommunicatorVisa import CommunicatorVisa
from sentio_prober_control.Sentio.CommandGroups.AuxCommandGroup import AuxCommandGroup
from sentio_prober_control.Sentio.CommandGroups.CompensationCommandGroup import CompensationCommandGroup
from sentio_prober_control.Sentio.CommandGroups.LoaderCommandGroup import LoaderCommandGroup
from sentio_prober_control.Sentio.CommandGroups.ProbeCommandGroup import ProbeCommandGroup
from sentio_prober_control.Sentio.CommandGroups.QAlibriaCommandGroup import QAlibriaCommandGroup
from sentio_prober_control.Sentio.CommandGroups.ServiceCommandGroup import ServiceCommandGroup
from sentio_prober_control.Sentio.CommandGroups.SiPHCommandGroup import SiPHCommandGroup
from sentio_prober_control.Sentio.CommandGroups.StatusCommandGroup import StatusCommandGroup
from sentio_prober_control.Sentio.CommandGroups.VisionCommandGroup import VisionCommandGroup
from sentio_prober_control.Sentio.CommandGroups.WafermapCommandGroup import WafermapCommandGroup
from sentio_prober_control.Sentio.Enumerations import ChuckSite


class SentioCommunicationType(Enum):
    """This enum defines different types of prober communication.

    Attributes:
        BackToContact (0): Chuck will move back to contact position after stepping.
        StepToSeparation (1): Chuck will move to separation position after stepping.
        LockContact (2): Chuck cannot step when at contact. You will have to manually move it away from its contact position before issuing the next step command.
    """

    TcpIp = 0
    Gpib = 1
    Visa = 2


class SentioProber(ProberBase):
    """This class represents the SENTIO probe station in python.
    It provides wrapper for most of the remote commands exposed by SENTIO.

    Attributes:
        aux (AuxCommandGroup): The aux command group provides access the the aux site modules functionality.
        loader (LoaderCommandGroup): The loader command group provides access to the loader modules functionality.
        map (WafermapCommandGroup): The wafermap command group provides access to the wafermap modules functionality.
        probe (ProbeCommandGroup): The probe command group provides access to the probe modules functionality.
        qalibria (QAlibriaCommandGroup): The qalibria command group provides access to the qalibria modules functionality.
        service (ServiceCommandGroup): The service command group provides access to the service modules functionality.
        siph (SiPHCommandGroup): The siph command group provides access to the SiPH modules functionality.
        status (StatusCommandGroup): The status command group provides access to the dashboard modules functionality. (formerly called status module)
        vision (VisionCommandGroup): The vision command group provides access to the vision modules functionality.
    """

    def __init__(self, comm: CommunicatorBase):
        """Construct a SENTIO prober object.

        The prober must be initialized with a communication object that
        specifies how the system communicates with the probe station.

        Args:
            comm (CommunicatorBase): The communicator to use for communication with the prober.
        """
        ProberBase.__init__(self, comm)

        self.__name = "SentioProber"
        self.comm.send("*RCS 1")  # switch to the native SENTIO remote command set

        self.aux: AuxCommandGroup = AuxCommandGroup(self)
        self.compensation: CompensationCommandGroup = CompensationCommandGroup(self)
        self.loader: LoaderCommandGroup = LoaderCommandGroup(self)
        self.map: WafermapCommandGroup = WafermapCommandGroup(self)
        self.probe: ProbeCommandGroup = ProbeCommandGroup(self)
        self.qalibria: QAlibriaCommandGroup = QAlibriaCommandGroup(self)
        self.service: ServiceCommandGroup = ServiceCommandGroup(self)
        self.siph: SiPHCommandGroup = SiPHCommandGroup(self)
        self.status: StatusCommandGroup = StatusCommandGroup(self)
        self.vision: VisionCommandGroup = VisionCommandGroup(self)


    def abort_command(self, cmd_id: int) -> Response:
        """Stop an ongoing asynchronous remote command.

        Args:
            cmd_id (int): The id of the async command to abort.

        Returns:
            A response object with the result of the command.
        """
        self.comm.send("abort_command {0}".format(cmd_id))
        return Response.check_resp(self.comm.read_line())


    def clear_contact(self, site: ChuckSite | None = None) -> Response:
        """Clear contact positions.

        Args:
            site (ChuckSite): The chuck site to clear. If None is given all sites will be cleared.

        Returns:
            A response object with the result of the command.
        """

        if site is None:
            self.comm.send("clear_contact")
        else:
            self.comm.send(f"clear_contact {site.toSentioAbbr()}")

        return Response.check_resp(self.comm.read_line())


    @staticmethod
    def create_prober(comm_type : str | SentioCommunicationType, arg1 : str | GpibCardVendor = "127.0.0.1:35555", arg2 : str = "") -> 'SentioProber':
        """ Create an instance of a SentioProber object that is bound to a certain communication method. Your choices of communication are tcpip, gpib and visa.
         
            Args:
                comm_type (str): The type of communication to use. Valid values are "tcpip", "gpib" and "visa".
                arg1 (object): For tcpip this is a single string specifying address and port 
                               like "127.0.0.1:35555". For visa this is the address of the device like "GPIB0::20::INSTR". For gpib two parameters are needed. 
                               The first one is the [type of driver/card](/Communication/GpibCardVendor) installed in the system, the second parameter is the address of the device like "GPIB0:20".
                arg2 (str): Only used for gpib communication. This is the GPIB address of the prober i.e. "GPIB0:20". 
         """
        
        if comm_type == "tcpip" or comm_type == SentioCommunicationType.TcpIp:
            return SentioProber(CommunicatorTcpIp.create(arg1))
        elif comm_type == "gpib" or comm_type == SentioCommunicationType.Gpib:
            return SentioProber(CommunicatorGpib.create(arg1, arg2))
        elif comm_type == "visa" or comm_type == SentioCommunicationType.Visa:
            return SentioProber(CommunicatorVisa.create(arg1))
        else:
            raise ValueError(f'Unknown prober type: "{comm_type}"')


    def enable_chuck_overtravel(self, stat: bool) -> Response:
        """Enable chuck overtravel.

        This function wraps SENTIO's "enable_chuck_overtravel" remote command.

        Args:
            stat (bool): True to enable, False to disable.

        Returns:
            A response object with the result of the command.
        """

        self.comm.send(f"enable_chuck_overtravel {stat}")
        return Response.check_resp(self.comm.read_line())


    def enable_chuck_hover(self, stat: bool) -> Response:
        """Enable chuck hover height.

        The Hover height is a height that is significantly closer to the chuck compared to the separation height.
        It is closer to the wafer but it is not safe for fast or long chuck moves as the chuck may be slightly
        tilted.

        This function wraps SENTIO's "enable_chuck_hover" remote command.

        Args:
            stat (bool): True to enable, False to disable.

        Returns:
            A response object with the result of the command.
        """

        self.comm.send(f"enable_chuck_hover {stat}")
        return Response.check_resp(self.comm.read_line())


    def enable_chuck_site_hover(self, site: ChuckSite, stat: bool) -> Response:
        """Enable chuck site hover height.

        The Hover height is a height that is significantly closer to the chuck compared to the separation height.
        It is closer to the wafer but it is not safe for fast or long chuck moves as the chuck may be slightly
        tilted.

        Args:
            site (ChuckSite): The chuck site to enable hover height for.
            stat (bool): True to enable, False to disable.

        Returns:
            A response object with the result of the command.
        """

        self.comm.send(f"enable_chuck_site_hover {site.toSentioAbbr()}, {stat}")
        return Response.check_resp(self.comm.read_line())


    def enable_chuck_site_overtravel(self, site: ChuckSite, stat: bool) -> Response:
        """Enable overtravel distance for a specific chuck site.

        Args:
            site (ChuckSite): The chuck site to enable overtravel distance for.
            stat (bool): True to enable, False to disable.

        Returns:
            A response object with the result of the command.
        """

        self.comm.send(f"enable_chuck_site_hover {site.toSentioAbbr()}, {stat}")
        return Response.check_resp(self.comm.read_line())


    def file_transfer(self, source: str, dest: str) -> Response:
        """Transfer a file to the prober.

        This function will transfer a file to the prober. The file will be stored in the position specified by
        the dest argument. Transmission of the file may take some time.

        Args:
            source (str): The path to the file to transfer.
            dest (str): The destination path on the prober. Must be a complete path including file name. Make sure that SENTIO has write access to the given destination.
        """
        # open file and encode with base64
        if not os.path.isfile(source):
            raise ProberException(f"File {source} not found!")

        file_bytes = open(source, "rb").read()
        encoded = base64.b64encode(file_bytes).decode("ascii")

        self.comm.send(f"file_transfer {dest}, {encoded}")
        return Response.check_resp(self.comm.read_line())


    def get_chuck_site_height(self, site: ChuckSite) -> Tuple[float, float, float, float]:
        """Retrieves height information of a chuck site

        Example:

            contact, separation, overtravel_dist, hover_gap = get_chuck_site_height(ChuckSite.Wafer)

        Gets for chuck site "Wafer" contact height, separation heights ,overtravel distance and hover height

        Args:
            site (ChuckSite): The chuck site to query.

        Returns:
            contact (float): contact height
            separation (float): separation gap
            overtravel_dist (float): overtravel distance
            hover_gap (float): hover gap
        """
        self.comm.send("get_chuck_site_heights {0}".format(site.toSentioAbbr()))
        resp = Response.check_resp(self.comm.read_line())

        tok = resp.message().split(",")

        contact = float(tok[0])
        separation = float(tok[1])
        overtravel_gap = float(tok[2])
        hover_gap = float(tok[3])

        return contact, separation, overtravel_gap, hover_gap


    def get_chuck_site_status(self, site: ChuckSite) -> Tuple[bool, bool, bool, bool]:
        """Get status of a chuck site.

        Wraps SENTIO's "get_chuck_site_status" remote command.

        Args:
            site (ChuckSite): The chuck site to query.

        Returns:
            hasHome (bool): True if the chuck site has a home position.
            hasContact (bool): True if the chuck site has a contact position.
            overtravelActive (bool): True if the overtravel is active.
            vacuumOn (bool): True if the vacuum is on.
        """
        self.comm.send("get_chuck_site_status {0}".format(site.toSentioAbbr()))
        resp = Response.check_resp(self.comm.read_line())

        tok = resp.message().split(",")

        def str_to_bool(v: str):
            if v == "0":
                return False
            elif v == "1":
                return True
            else:
                raise ProberException(f"get_chuck_site_status: Invalid boolean value ({v})")

        hasHome = str_to_bool(tok[0])
        hasContact = str_to_bool(tok[1])
        overtravelActive = str_to_bool(tok[2])
        vacuumOn = str_to_bool(tok[3])

        return hasHome, hasContact, overtravelActive, vacuumOn


    def get_chuck_theta(self, site: ChuckSite) -> float:
        """Get the current angle of the chuck.

        Wraps SENTIO's "get_chuck_theta" remote command.

        Args:
            site (ChuckSite): The chuck site to query.

        Raises:
            ProberException: If no available port is found.

        Returns:
            angle (float): The current angle of the chuck site in degrees.
        """

        self.comm.send("get_chuck_theta {0}".format(site.toSentioAbbr()))
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())


    def get_chuck_xy(self, site: ChuckSite, ref: ChuckXYReference) -> Tuple[float, float]:
        """Get current chuck xy position with respect to a given reference.

        Args:
            site (ChuckSite): The chuck site to query.
            ref (ChuckXYReference): The reference to use for the query.

        Returns:
            x (float): x position in micrometer.
            y (float): y position in micrometer.
        """

        if site is None:
            self.comm.send(f"get_chuck_xy {site.toSentioAbbr()}")
        else:
            self.comm.send(f"get_chuck_xy {site.toSentioAbbr()}, {ref.toSentioAbbr()}")

        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1])


    @deprecated(reason="Duplicate functionality; Use get_chuck_xy instead")
    def get_chuck_xy_pos(self) -> Tuple[float, float]:
        self.comm.send("get_chuck_xy")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        curX = float(tok[0])
        curY = float(tok[1])
        return curX, curY


    def get_chuck_z(self, ref: ChuckZReference) -> float:
        """Get chuck z position.

        Args:
            ref (ChuckZReference): The reference to use for the query.

        Returns:
            height (float): The actual z position of the chuck in micrometer (from axis zero).
        """

        self.comm.send(f"get_chuck_z {ref.toSentioAbbr()}")
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())

    def get_project(self, pfi: ProjectFileInfo = ProjectFileInfo.FullPath) -> str:
        """Get the name of the current project.

        Args:
            pfi (ProjectFileInfo): The type of information to get.

        Returns:
            project_name (str): The name of the current project.
        """

        self.comm.send(f"get_project {pfi.toSentioAbbr()}")
        resp = Response.check_resp(self.comm.read_line())
        return resp.message()

    def get_scope_xy(self) -> Tuple[float, float]:
        """Get current scope xy position.

        The returned position is an absolute position with respect to the axis zero in micrometer.

        Returns:
            x (float): The current x position in micrometer.
            y (float): The current y position in micrometer.
        """

        self.comm.send("get_scope_xy")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1])

    def get_scope_z(self) -> float:
        """Get scope z position in micrometer from axis zero.

        Returns:
            height (float): The z position in micrometer.
        """

        self.comm.send("get_scope_z")
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())


    def get_scope_site(self, idx : int) -> Tuple[str, float, float, bool]:
        """ Get scope site data.

            This command queries the name and position of a scope site. The index is zero based.

        Args:
            idx (int): The index of the site to query.

        Returns:
            num (int): The number of defined scope sites.
        """

        self.comm.send(f"get_scope_site {idx}")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")

        id = tok[0]
        x = float(tok[1])
        y = float(tok[2])

        # tok[3] is the reference which should always be "home"

        enabled = tok[4].upper() == "TRUE"

        return (id, x, y, enabled)
    

    def get_scope_site_num(self) -> int:
        """Get total number of scope sites.

        The scope stage can store a list of sites. Each site has a position and a name.
        This function returns the number of defined scope sites.

        Returns:
            num (int): The number of defined scope sites.
        """

        self.comm.send("get_scope_site_num")
        resp = Response.check_resp(self.comm.read_line())
        return int(resp.message())


    def has_chuck_xyz(self) -> bool:
        """Returns True if the chuck has xyz axes.

        Returns:
            has_xyz (bool): True if the chuck has xyz axes.
        """

        self.comm.send("has_chuck_xyz")
        resp = Response.check_resp(self.comm.read_line())
        return resp.message().upper() == "YES"


    def has_scope_xyz(self) -> bool:
        """Returns True if the scope has xyz axes.

        Returns:
            has_xyz (bool): True if the scope has xyz axes.
        """

        self.comm.send("has_scope_xyz")
        resp = Response.check_resp(self.comm.read_line())
        return resp.message().upper() == "YES"


    def has_scope_z(self) -> bool:
        """Returns true if the microscope has a motorized z axis.

        Returns:
            True if the scope has z axes.
        """

        self.comm.send("has_scope_z")
        resp = Response.check_resp(self.comm.read_line())
        return resp.message().upper() == "YES"


    def initialize_if_needed(self):
        """Initialize the prober if it is not already initialized.

        This command will check if the prober is already initialized. If not
        it will start the initialization process and wait for it to complete.

        You do not have to call waitcomplete after this function on your own!
        """
        isInitialized, isMeasuring, isLoaderBusy = self.status.get_machine_status()

        if not isInitialized:
            resp = self.start_initialization()

            if not resp.ok():
                raise ProberException(f"Cannot start initialization: {resp.message()}")

            resp = self.wait_complete(resp.cmd_id(), 180)
            if not resp.ok():
                raise ProberException(f"Initialization failed: {resp.message()}")


    def local_mode(self):
        """Switch the prober back into local mode.

        The probe station will automatically enter remote mode when a remote command is received.
        It will remian in remote mode even after the script is finished. This command can be used
        to switch the machine back into local mode and thus enable its UI.
        """
        self.comm.send("*LOCAL")


    def move_chuck_contact(self) -> float:
        """Move the chuck to contact height.

        Wraps SENTIO's "move_chuck_contact" remote command.

        Returns:
            height (float): The contact height in micrometer from chuck z axis zero.
        """

        self.comm.send("move_chuck_contact")
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())


    def move_chuck_home(self) -> Tuple[float, float]:
        """Move chuck to its home position.

        Returns:
            x (float): The x position in micrometer.
            y (float): The y position in micrometer.
        """

        self.comm.send("move_chuck_home ")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1])


    def move_chuck_load(self, pos: LoadPosition) -> Response:
        """Move chuck to load position.

        Wraps SENTIO's "move_chuck_load" remote command.

        Args:
            pos (LoadPosition): The load position to move to.

        Returns:
            A response object with the result of the command.
        """
        self.comm.send("move_chuck_load {0}".format(pos.toSentioAbbr()))
        return Response.check_resp(self.comm.read_line())


    def move_chuck_separation(self) -> float:
        """Move the chuck to separation height.

        Returns:
            The separation height in micrometer from chuck z axis zero.
        """
        self.comm.send("move_chuck_separation ")
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())


    def move_chuck_site(self, site: ChuckSite) -> Tuple[float, float, float, float]:
        """Moves chuck to the last active position of the selected chuck site.

        Wraps SENTIO's "move_chuck_site" remote command.

        Returns:
            x (float): The x position in micrometer.
            y (float): The y position in micrometer.
            z (float): The z height in micrometer.
            theta (float): The theta angle in degrees.
        """
        self.comm.send("move_chuck_site {0}".format(site.toSentioAbbr()))
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1]), float(tok[2]), float(tok[3])


    def move_chuck_theta(self, ref: ChuckThetaReference, angle: float) -> float:
        """Move chuck theta axis to a given angle.

        Wraps SENTIO's "move_chuck_theta" remote command.

        Args:
            ref: The reference to use for the move.
            angle: The angle to move to in degrees.
        """
        self.comm.send("move_chuck_theta {0}, {1}".format(ref.toSentioAbbr(), angle))
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())


    def move_chuck_xy(self, ref: ChuckXYReference, x: float, y: float) -> Tuple[float, float]:
        """Move chuck to a given xy position.

        Wraps SENTIO's "move_chuck_xy" remote command.

        Args:
            ref: The reference to use for the move.
            x: The x position to move to in micrometer.
            y: The y position to move to in micrometer.

        Returns:
            x: The chuck x position after the move in micrometer (from zero)
            y: The chuck y position after the move in micrometer (from zero)
        """
        self.comm.send("move_chuck_xy {0}, {1}, {2}".format(ref.toSentioAbbr(), x, y))
        resp = Response.check_resp(self.comm.read_line())

        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1])

    def move_chuck_z(self, ref: ChuckZReference, z: float) -> float:
        """Move chuck to a given z position.

        Wraps SENTIO's "move_chuck_z" remote command.

        Args:
            ref: The z-reference to use for the move.
            z: The z position to move to in micrometer.

        Returns:
            The actual z position in micrometer after the move.
        """
        self.comm.send("move_chuck_z {0}, {1}".format(ref.toSentioAbbr(), z))
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())

    def move_chuck_work_area(self, area: WorkArea) -> Response:
        """Move the chuck to a given work area.

        A SENTIO probe station can have different work areas. One area is for probing. This is the default
        work area and present on all probe stations. When the chuck is at probing position is is roughly
        centered with respect to the platen and located in a space were probes can go into contact.
        Depending on the machine type a prober may be equipped with a second work area. This
        area is located under an optional off-axis camera. The off-axis camera allows
        inspection of the wafer when the scope camera view is blocked by certain probe card types.

        This function wraps SENTIO's "move_chuck_work_area" remote command.

        Args:
            area: The work area to move to.

        Returns:
            A response object with the result of the command.
        """
        self.comm.send(f"move_chuck_work_area {area.toSentioAbbr()}")
        return Response.check_resp(self.comm.read_line())

    def move_scope_xy(
        self, ref: ScopeXYReference, x: float, y: float
    ) -> Tuple[float, float]:
        """Move scope to a given xy position.

        Args:
            ref: The reference to use for the move.
            x: The x position to move to in micrometer.
            y: The y position to move to in micrometer.

        Returns:
            x: Scope x position after the move in micrometers (from zero)
            y: Scope x position after the move in micrometers (from zero)
        """

        self.comm.send(f"move_scope_xy {ref.toSentioAbbr()}, {x}, {y}")
        resp = Response.check_resp(self.comm.read_line())

        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1])

    def move_scope_lift(self, state: bool):
        """Move scope to its lift position.

        The scope lift position is a position where the scope is
        at its axis maximum. This position will give you the maximum
        possible are of unhindered operation when changing probe cards
        or other maintenance tasks.

        Args:
            state: True to move to the lift position, False to move away from the lift position.
        """
        self.comm.send(f"move_scope_lift {state}")
        Response.check_resp(self.comm.read_line())

    def move_scope_z(self, ref: ScopeZReference, z: float) -> float:
        """Move scope to a given z position.

        Args:
            ref: The reference to use for the move.
            z: The z position to move to in micrometer.

        Returns:
            The actual z position in micrometer after the move.
        """
        self.comm.send("move_scope_z {0}, {1}".format(ref.toSentioAbbr(), z))
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())

    def move_vce_z(self, stage: Stage, ref: VceZReference, z: float) -> float:
        """Move VCE stage to a given z position.

        Args:
            ref: The reference to use for the move.
            z: The z position to move to in micrometer.

        Returns:
            The actual z position in micrometer after the move.
        """
        if stage != Stage.Vce and stage != Stage.Vce2:
            raise ProberException(
                f"This command can only be applied to vce stages! (stage={0})"
            )

        self.comm.send(f"move_vce_z {stage.toSentioAbbr()}, {ref.toSentioAbbr()}, {z}")
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())

    def name(self) -> str:
        """Returns the name of the prober.

        Returns:
            This function will always return the string "SentioProber".
        """
        return self.__name

    def open_project(self, project: str, restore_heights: bool = False):
        """Open a SENTIO project file.

        Wraps SENTIO's "open_project" remote command.

        Args:
            project: The name or path of the project to open. If a full
                     path to the trex project file is given SENTIO will try to open this file.
                     If the argument does not contain a path SENTIO will look in its default
                     project folder for a matching project and open it.
            restore_heights: If set to true SENTIO will restore the contact
                             heights from the project. Be carefull when using this option because
                             the contact heights may have been become invalid since creating the project
                             due to a probe card change.
        """
        self.comm.send(f"open_project {project}, {restore_heights}")
        Response.check_resp(self.comm.read_line())

    def query_command_status(self, cmd_id: int) -> Response:
        """Query the status of an async command.

        This command will send a query to the prober to get the status
        of an async command. The submitted command id must be a valid id
        of an async command that was previously started at the prober.

        When an async command is being executed SENTIO can receive other
        remnote commands. This command is intended to be used in a polling
        loop to query the status of ongoing remote commands.

        Example:

        <pre><code>while True:
            time.sleep(1)
            resp = prober.query_command_status(cmd_id)
            if (resp.errc()!=RemoteCommandError.CommandPending):
                break;</code></pre>

        Args:
            cmd_id: The id of the async command to query.

        Returns:
            A response object with the result of the command.
        """
        self.comm.send("query_command_status {0}".format(cmd_id))
        resp = Response.parse_resp(self.comm.read_line())
        return resp

    def save_config(self):
        """Save the SENTIO configuration file.

        Wraps SENTIO's "save_config" remote command.
        """
        self.comm.send("save_config")
        Response.check_resp(self.comm.read_line())

    def save_project(self, project: str):
        """Save the current SENTIO project.

        Wraps SENTIO's "save_project" remote command.
        """
        self.comm.send("save_project " + project)
        Response.check_resp(self.comm.read_line())


    def select_module(self, module: Module, tabSheet: str | None = None):
        """Activate a given SENTIO module.

        In response to this function SENTIO will switch its user interface to make
        the given module the active one.

        This function wraps the "select_module" remote command of SENTIO.

        Args:
            module: The module to activate.
            tabSheet: The name of the module tab to activate. If None is given the default tab will be activated.
        """
        if tabSheet is not None:
            self.comm.send(f"select_module {module.toSentioAbbr()}, {tabSheet}")
        else:
            self.comm.send(f"select_module {module.toSentioAbbr()}")

        Response.check_resp(self.comm.read_line())


    def send_cmd(self, cmd: str) -> Response:
        """Sends a command to the prober and return a response object.

        This function is intended for directly sending remote commands that
        are not yet included in the python wrapper. It will send the command
        and parse the respone from SENTIO.

        Do NOT send low level commands that do not have a response (i.e. "*LOCAL").
        This will lock the communication pipeline as it is waiting for a
        response that never arrives.

        It will then return a Response object with the extracted data from
        SENTIO's response.

        Returns:
            A response object with the result of the command.
        """
        self.comm.send(cmd)
        return Response.check_resp(self.comm.read_line())

    def set_chuck_site_height(
        self,
        site: ChuckSite,
        contact: float,
        separation: float,
        overtravel_dist: float,
        hover_gap: float,
    ) -> Response:
        """Sets z position information of a chuck site

        Example:

            prober.set_chuck_site_height(ChuckSite.Wafer,16000,250,20,50)

        Will set the contact height of the wafer site to 16000 µm with a separation height of 250 µm an overtravel of 20 and a hover height of 50

        Args:
            site: The chuck site to query.
            contact: The new contact height in micrometer.
            separation: The new separation height in micrometer.
            overtravel_dist: The new overtravel distance in micrometer.
            hover_gap: The new hover gap in micrometer.
        """
        par: str = f"{site.toSentioAbbr()},{contact},{separation},{overtravel_dist},{hover_gap}"
        self.comm.send("set_chuck_site_heights {0}".format(par))
        return Response.check_resp(self.comm.read_line())


    def set_ink(self, idx_inker : int) -> None:
        """ Trigger the inker device.

            Wraps SENTIO's "set_ink" remote command.

        Args:
            idx_inker: The index of the inker to trigger. The index is one based (use either 1 or 2). 
        """

        self.comm.send(f"set_ink {idx_inker}")
        Response.check_resp(self.comm.read_line())


    def set_stepping_contact_mode(self, mode: SteppingContactMode) -> None:
        """Change the stepping contact mode.

        The stepping contact mode defines what happens during stepping over a wafer.
        The following modes are available:

        * BackToContact: The chuck will step into contact position
        * StepToSeparation: The chuck will step into separation position. You have to move into contact with a seaparate command.
        * LockContack: The Chuck is not allowed to leave contact position automatically as part of a stepping command. You have to move into separation with a separate command bevore being able to step to the next die..

        Args:
             mode: The stepping contact mode to set.
        """

        self.comm.send(f"set_stepping_contact_mode {mode.toSentioAbbr()}")
        Response.check_resp(self.comm.read_line())


    def set_vacuum(self, site: ChuckSite, stat: bool) -> Response:
        """Switches the vacuum of a chuck site on or off.

        Args:
            site: The chuck site to switch the vacuum for.
            stat: True to switch the vacuum on, False to switch it off.

        Returns:
            A response object with the result of the command.
        """

        self.comm.send(f"set_vacuum {site.toSentioAbbr()}, {stat}")
        return Response.check_resp(self.comm.read_line())


    def show_hint(self, msg: str, subtext: str) -> Response:
        """Show an on screen message (hint) and return immediately

        Hints are on screen messages that pop up in SENTIO's lower left corner. This hint will
        disappears automatically after a few seconds.

        This function wraps SENTIO's "status:show_hint" remote command.

        Args:
            msg: The message to display.
            subtext: The subtext to display.
        """
        self.comm.send(f'status:show_hint "{msg}", "{subtext}"')
        return Response.check_resp(self.comm.read_line())


    def show_hint_and_wait(
        self,
        msg: str,
        subtext: str,
        button_caption: str,
        timeout: int = 180,
        lock_ui: bool = True,
    ) -> None:
        """Show an on screen message with a button wait for a button to be pressed.

        Hints pop up in SENTIO's lower left corner. This function will display a hint with a button
        and only return once the button has been pressed.

        Args:
            msg: The message to display.
            subtext: The subtext to display. Subtext is displayed in a second line with a slightly smaller font.
            button_caption: The caption of the button.
            timeout: An optional timeout in seconds after which the dialog will be closed automatically. (default = 180 s)
            lock_ui: An optional flag that determines wether the UI shall be locked. Most of the UI is disabled in
                     remote mode anyway. This button affects only the on screen interactions on the right side of the main
                     module view. (default = True)
        """

        self.comm.send(
            f'status:start_show_hint "{msg}", "{subtext}", "{button_caption}", "{lock_ui}"'
        )
        resp = Response.check_resp(self.comm.read_line())

        # wait for button press
        self.comm.send(f"wait_complete {resp.cmd_id()}, {timeout}")
        Response.check_resp(self.comm.read_line())


    def show_message(
        self, msg: str, buttons: DialogButtons, caption: str, dialog_timeout: int = 180
    ) -> DialogButtons:
        """Pop up a message dialog in SENTIO and wait for the result.

        Args:
            msg: The message to display.
            buttons: The buttons to display.
            caption: The caption of the message box.
            dialog_timeout: An optional dialog timeout in seconds after which the dialog will be closed automatically.

        Returns:
            The button that was as an DialogButtons enum value.
        """
        self.comm.send(
            "status:start_show_message {0}, {1}, {2}".format(
                msg, buttons.toSentioAbbr(), caption
            )
        )
        resp = Response.check_resp(self.comm.read_line())

        # wait for button press
        self.comm.send("wait_complete {0}, {1}".format(resp.cmd_id(), dialog_timeout))
        resp = Response.check_resp(self.comm.read_line())

        if resp.message().lower() == "ok":
            return DialogButtons.Ok

        if resp.message().lower() == "yes":
            return DialogButtons.Yes

        if resp.message().lower() == "no":
            return DialogButtons.No

        if resp.message().lower() == "cancel":
            return DialogButtons.Cancel

        raise ProberException("Invalid dialog button return value")


    def start_initialization(self) -> Response:
        """Start the initialization of the probe station.

        This function will start the initialization of the prober. This is an
        async command that will return immediately.

        The initialization process will take some time to complete. Use
        waitcomplete to wait for the initialization to complete.
        """
        self.comm.send("start_initialization")
        return Response.check_resp(self.comm.read_line())


    def wait_all(self, timeout: int = 90) -> Response:
        """Wait until all async commands have finished.

        Args:
            timeout: The timeout in seconds.

        Returns:
            A response object with the result of the command.
        """
        self.comm.send(f"wait_all {timeout}")
        return Response.check_resp(self.comm.read_line())


    def wait_complete(self, id_or_resp: int | Response, timeout: int = 300) -> Response:
        """Wait for a single async command to complete.

        Args:
            cmd_id: The id of the async command to wait for.
            timeout: The timeout in seconds.

        Returns:
            A response object with the result of the command.
        """
        if isinstance(id_or_resp, Response):
            self.comm.send(f"wait_complete {id_or_resp.cmd_id()}, {timeout}")    
        else:
            self.comm.send(f"wait_complete {id_or_resp}, {timeout}")

        return Response.check_resp(self.comm.read_line())



import re

from typing import List, Tuple

from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.Helper import Helper
from sentio_prober_control.Sentio.Enumerations import VirtualCarrierInitFlags, VirtualCarrierStepProcessingState, LoaderStation


class LoaderVirtualCarrierCommandGroup(CommandGroupBase):
    """A command group for the Virtual Carrier functionality.

    You are not meant to instantiate this class directly. Access it via the loader attribute
    of the [SentioProber](SentioProber.md) class.
    """

    def __init__(self, parent : CommandGroupBase) -> None:
        super().__init__(parent)


    def list(self) -> List[str]:
        """List all virtual carriers defined in SENTIO.

        Wraps Sentios "loader:vc:list" remote command.

        Returns:
            carriers (Tuple[str]): A tuple with the names of the virtual carriers.
        """

        try:
            self.comm.send("loader:vc:list")
            resp = Response.check_resp(self.comm.read_line())
            
            list : List[str] = resp.message().split(',')
            return list
        except:
            return []
        

    def select(self, vc_name : str) -> Response:
        """Select a virtual carrier.

        Wraps Sentios "loader:vc:select" remote command.

        Args:
            vc_name (str): The name of the virtual carrier to select.

        Returns:
            response (Response): A Response object.
        """

        self.comm.send(f"loader:vc:select {vc_name}")
        return Response.check_resp(self.comm.read_line())


    def start_load_first(self, cleanup : bool = False) -> Response:
        """Start loading the first wafer in the selected virtual carrier.
            
            This is an async command. You need to wait for the command to 
            complete.

        Wraps Sentios "loader:vc:start_load_first" remote command.

        Returns:
            resp (Response): A response object.
        """

        self.comm.send(f"loader:vc:start_load_first {cleanup}")
        resp = Response.check_resp(self.comm.read_line())
        return resp


#    def __split_string(self, s : str) -> List[str]:
#        matches = re.findall(r'"[^"]*"|[^,]+', s)
#        return [match.strip('"') for match in matches]

    def initialize(self, carrier_name : str, mode : VirtualCarrierInitFlags = VirtualCarrierInitFlags.Start, timeout : int = 90) -> List[Tuple[VirtualCarrierStepProcessingState, str, LoaderStation, int, float, int]]:
        """Initializes the selected virtual carrier.
        
            Wraps Sentios "loader:vc:start_initialize" remote command. This command synchronizes the async
            start_initialize command with a wait_complete command.

            Args:
                carrier_name (str): The name of the virtual carrier to initialize. This carrier will be made SENTIO's active carrier.
                continue_process (bool): A boolean flag to indicate if a previous virtual carrier measurement shall be continued. (default is False)
                timeout (int): The timeout in seconds. (default is 90 seconds)

            Returns:
                A List of steps to execute.
        """

        resp : Response = self.prober.send_cmd(f"loader:vc:start_initialize {carrier_name}, {mode.toSentioAbbr()}")
        resp = self.prober.wait_complete(resp.cmd_id())

        # parse processing steps into a list of tuples
        steps : List[Tuple[VirtualCarrierStepProcessingState, str, LoaderStation, int, float, int]] = []

        steps_as_str : List[str] = Helper.split_string(resp.message(), ',')
#        steps_as_str : List[str] = resp.message().split(',')

        for it in steps_as_str:
#            col = it.split(' ')
            col = Helper.split_string(it, ' ')

            state = VirtualCarrierStepProcessingState[col[0]]
            id = col[1]
            station = LoaderStation[col[2]]
            slot = int(col[3])
            temp = float(col[4])
            probecard_idx = int(col[5])
            steps.append((state, id, station, slot, temp, probecard_idx))

        return steps


    def next_step(self, timeout : int = 90) -> Tuple[VirtualCarrierStepProcessingState, str, LoaderStation, int, float, int]:
        """Execute the next processing step of the virtual carrier.

            Wraps Sentios "loader:vc:start_next_step" remote command.

        Returns:
            A tuple with the processing step information.
        """

        resp : Response = self.prober.send_cmd("loader:vc:start_next_step")
        resp = self.prober.wait_complete(resp)

        col = resp.message().split(',')
        state = VirtualCarrierStepProcessingState[col[0]]
        id = col[1]
        station = LoaderStation[col[2]]
        slot = int(col[3])
        temp = float(col[4])
        probecard_idx = int(col[5])

        return (state, id, station, slot, temp, probecard_idx)




from sentio_prober_control.Sentio.Enumerations import PoiReferenceXy, Stage
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase


class WafermapPoiCommandGroup(CommandGroupBase):
    """A command group for working with Points of Interest (POI) on the wafermap."""

    def add(self, x: float, y: float, desc: str) -> None:
        """Add a POI to the list.

        Wraps SENTIO's map:poi:add remote command.

        Args:
            x: The x coordinate of the POI.
            y: The y coordinate of the POI.
            desc: The description of the POI.
        """
        self.comm.send(f"map:poi:add {x}, {y}, {desc}")
        Response.check_resp(self.comm.read_line())


    def get_num(self) -> int:
        """Returns the number of POIs in the list.

        Wraps SENTIO's map:poi:get_num remote command.

        Returns:
            The number of POIs in the list.
        """
        self.comm.send("map:poi:get_num")
        resp = Response.check_resp(self.comm.read_line())
        return int(resp.message())


    def reset(self, stage: Stage, refXy: PoiReferenceXy) -> None:
        """Reset the list of POIs.

        Clears the list and resets its stage and position reference settings.

        Wraps SENTIO's map:poi:reset remote command.

        Args:
            stage: The stage to reset the POIs for.
            refXy: The reference point for the POIs.
        """
        self.comm.send("map:poi:reset {0}, {1}".format(stage.toSentioAbbr(), refXy.toSentioAbbr()))
        Response.check_resp(self.comm.read_line())


    def step(self, target: str | int) -> None:
        """Step to a POI in the list.

        Wraps SENTIO's map:poi:step remote command.

        Args:
            target: The target POI to step to. This is either the index of the poi or the id of the poi.
        """
        self.comm.send(f"map:poi:step {target}")
        Response.check_resp(self.comm.read_line())


    def step_first(self) -> None:
        """Step to the first POI in the list.

        Wraps SENTIO's map:poi:step_first remote command.
        """
        self.comm.send("map:poi:step_first")
        Response.check_resp(self.comm.read_line())


    def step_next(self) -> None:
        """Step to the next POI in the list.

        Wrap SENTIO's map:poi:step_next remote command.
        """
        self.comm.send("map:poi:step_next")
        Response.check_resp(self.comm.read_line())



""" This module contains low level device drivers for the MPI probe station.

    These classes provide an interface to the native GPIB drivers from ADLINK and NI. There is no documentation
    available since this in an internal implementation detail.

    @private
"""



from typing import Tuple

from sentio_prober_control.Sentio.Enumerations import CompensationMode, CompensationType
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase

from deprecated import deprecated


class VisionCompensationGroup(CommandGroupBase):
    """This command group contains functions for working with x,y and z compensation.

    You are not meant to instantiate this class directly. Access it via the compensation attribute
    of the vision attribute of the [SentioProber](SentioProber.md) class.
    """

    def __init__(self, comm):
        super().__init__(comm)


    @deprecated("Use vision.compensation.enable() instead")
    def set_compensation(self, comp: CompensationMode, enable: bool) -> Tuple[str, str]:
        self.comm.send(f"vis:compensation:enable {comp.toSentioAbbr()}, {enable}")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return tok[0], tok[1]


    def enable(self, comp: CompensationMode, enable: bool) -> Tuple[str, str]:
        """Enable or disable compensation for a given subsystem.

        Wraps Sentios "vis:compensation:enable" command.

        Args:
            comp: The compensation to enable or disable.
            enable: True to enable, False to disable.

        Returns:
            XY-Mode: State of the XY compensation.
            Z-Mode: State of the Z compensation.
        """

        self.comm.send(f"vis:compensation:enable {comp.toSentioAbbr()}, {enable}")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return tok[0], tok[1]


    def start_execute(self, type: CompensationType, mode: CompensationMode) -> Response:
        """Start the execution of a compensation.

        Wraps Sentios "vis:compensation:start_execute" remote command.

        Args:
            type: The type of compensation to execute.
            mode: The mode of compensation to execute.

        Returns:
            A Response object.
        """

        self.comm.send(f"vis:compensation:start_execute {type.toSentioAbbr()}, {mode.toSentioAbbr()}")
        return Response.check_resp(self.comm.read_line())



import socket

from sentio_prober_control.Communication.CommunicatorBase import CommunicatorBase


class CommunicatorTcpIp(CommunicatorBase):
    """Communicator for TCP/IP communication."""

    def __init__(self):
        """Construcst a TCP/IP communicator."""
        self.__socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    @staticmethod
    def create(addr: str):
        """Create an instance of a TCP/IP communicator.

        :param addr: A string that specifies the address of the TCP/IP device to connect to. The address must have the format "IP_ADDRESS&colon;PORT"
        """
        c = CommunicatorTcpIp()
        c.connect(addr)
        return c

    def connect(self, address: str):
        """Connect to a TCP/IP device at the specified address.

        :param addr: A string that specifies the address of the TCP/IP device to connect to. The address must have the format "IP_ADDRESS&colon;PORT"
        """
        tok = address.split(":")
        self.__address = tok[0]
        self.__port = int(tok[1])

        if CommunicatorBase._verbose:
            print(
                "Connecting comunicator to {0}:{1}".format(self.__address, self.__port)
            )

        self.__socket.connect((self.__address, self.__port))

    def disconnect(self):
        """Disconnect from the TCP/IP device."""
        if CommunicatorBase._verbose:
            print("Diconnecting comunicator to {0}".format(self.__address))

        self.__socket.close()

    def send(self, msg: str):
        """Send a command to the TCP/IP device.

        :param msg: The command to send.
        """

        if CommunicatorBase._verbose:
            print('Sending "{0}"'.format(msg))

        self.__socket.send((msg + "\n").encode())

    def read_line(self):
        """Read a line from the TCP/IP device.

        :return: The read line.
        """
        return self.__socket.makefile().readline().rstrip()



from abc import ABC
from typing import Union

from sentio_prober_control.Communication.CommunicatorBase import CommunicatorBase


class CommandGroupBase(ABC):
    """Base class for all command groups.

    The functionality of the prober control suiote is organized in command groups.
    Each command group contains a set of commands that are logically related.
    """

    def __init__(self, parent : Union['CommandGroupBase', 'SentioProber']) -> None:
        self.__parent = parent


    @property
    def prober(self) -> 'SentioProber':
        """Get the prober object.

        Returns:
            prober (SentioProber): The prober object.
        """

        from sentio_prober_control.Sentio.ProberSentio import SentioProber

        if isinstance(self.__parent, SentioProber):
            return self.__parent
        else:
            return self.__parent.prober
        
       
    @property
    def comm(self) -> CommunicatorBase:
        """Get the communicator object.

        Returns:
            comm (CommunicatorBase): The communicator object.
        """
        return self.__parent.comm



from typing import Tuple

from sentio_prober_control.Sentio.Enumerations import BinSelection, BinQuality
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase


class WafermapBinsCommandGroup(CommandGroupBase):
    """This command group bundles functions for setting up and using the binning table of the wafermap."""

    def clear_all(self) -> None:
        """Clear all bins. Remove the bin code from all dies and sibsites."""
        self.comm.send("map:bins:clear_all")
        Response.check_resp(self.comm.read_line())


    def clear_all_values(self) -> None:
        """Removes all temporarily stored values from the dies.

        Each die can store a single floating point value. This value can be used for
        visualizing parameters across the wafer.
        """
        self.comm.send("map:bins:clear_all_values")
        Response.check_resp(self.comm.read_line())


    def get_bin(self, col: int | None = None, row: int | None = None, site: int | None = None) -> int:
        """Get the bin information of a die or a subsite.

        Wraps SENTIO's map:bins:get_bin remote command.

        Args:
            col: The column of the die.
            row: The row of the die.
            site: The site of the die.

        Returns:
            The bin value of the die or subsite.
        """
        if col is None and row is None and site is None:
            self.comm.send(f"map:bins:get_bin")
        elif site is None and col is not None and row is not None:
            self.comm.send(f"map:bins:get_bin {col}, {row}")
        elif site is not None and col is not None and row is not None:
            self.comm.send(f"map:bins:get_bin {col}, {row}, {site}")
        else:
            raise ValueError("get_bin command requires either no parameter or column and row or column, row and site.")
        
        resp = Response.check_resp(self.comm.read_line())
        return int(resp.message())


    def get_bin_info(self, bin : int) -> Tuple[int, str, BinQuality, str]:
        """Get the information of a bin code from the binning table defined in the wafer map.
        
            Wraps SENTIO's map:bins:get_bin_info remote command.

            Returns:
                A tuple containing the following data items: bin index (same value as the argument), bin description, bin quality (pass/fail information) and bin color.
        """
        self.comm.send(f"map:bins:get_bin_info {bin}")        
        resp = Response.check_resp(self.comm.read_line())

        values = resp.message().split(",")
        return int(values[0]), values[1], BinQuality[values[2]], values[3]
    

    def get_num_bins(self) -> int:
        """Get the number of bins in the binning table.
        
            Returns:
                The number of bins in the binning table.
        """
        self.comm.send("map:bins:get_num_bins")
        resp = Response.check_resp(self.comm.read_line())
        return int(resp.message())
    

    def load(self, file: str) -> None:
        """Load a binning table from file.

        Wraps SENTIO's map:bins:load remote command.

        Args:
            file: The file to load the binning table from.
        """
        self.comm.send(f"map:bins:load {file}")
        Response.check_resp(self.comm.read_line())


    def set_all(self, bin_val: int, selection: BinSelection) -> None:
        """Sets the bins of all dies on the wafermap to a specific value.

        Wraps SENTIO's map:bins:set_all remote command.

        Args:
            bin_val: The bin value to set.
            selection: The selection of dies to set the bin value for.
        """
        self.comm.send(f"map:bins:set_all {bin_val}, {selection.toSentioAbbr()}")
        Response.check_resp(self.comm.read_line())  #


    def set_bin(self, bin_value: int, col: int | None = None, row: int | None = None, site: int | None = None) -> None:
        """Set a single bin.

        Wraps SENTIO's map:bins:set_bin remote command.

        Args:
            bin_value: The bin value to set.
            col: The column of the die.#
            row: The row of the die.
            site: The site of the die.
        """
        if col is None and row is None and site is None:
            self.comm.send(f"map:bins:set_bin {bin_value}")
        elif site is None and col is not None and row is not None:
            self.comm.send(f"map:bins:set_bin {bin_value}, {col}, {row}")
        elif site is not None and col is not None and row is not None:
            self.comm.send(f"map:bins:set_bin {bin_value}, {col}, {row}, {site}")
        else:
            raise ValueError("set_bin command requires either no parameter or column and row or column, row and site.")

        Response.check_resp(self.comm.read_line())


    def set_value(self, value: float, col: int, row: int) -> None:
        """Set a value on a single die.

        Args:
            value: The value to set.
            col: The column of the die.
            row: The row of the die.
        """
        self.comm.send("map:bins:set_value {0}, {1}, {2}".format(value, col, row))
        Response.check_resp(self.comm.read_line())



from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.ProberBase import ProberException
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase


class AuxCleaningGroup(CommandGroupBase):
    """This command group provides functions for for cleaning the probe.
    You are not meant to create instances of this class on your own.
    Instead use the cleaning property of the AuxCommandGroup.

    Example:

    ```py
        from sentio_prober_control.Sentio.ProberSentio import SentioProber

        prober = SentioProber.create_prober("tcpip", "127.0.0.1:35555")
        prober.aux.cleaning.enable_auto(True)
    ```

    """

    def enable_auto(self, stat: bool) -> None:
        """Enable automatic probe cleaning.

        Args:
            stat (bool): A flag indicating whether to enable or disable cleaning.
        """

        self.comm.send(f"aux:cleaning:enable_auto {stat}")
        Response.check_resp(self.comm.read_line())


    def start(self, touchdowns: int | None = None) -> None:
        """Start the cleaning procedure.

        Args:
            touchdowns (int): The number of touchdowns to perform.
        """
        if touchdowns is None:
            self.comm.send(f"aux:cleaning:start")
        else:
            self.comm.send(f"aux:cleaning:start {touchdowns}")

        Response.check_resp(self.comm.read_line())



""" @private """

import ctypes
from ctypes import byref, c_byte, c_char_p, c_int, c_ushort, c_void_p, c_wchar_p


class GpibAdlinkDriver:
    """A python wrapper for the ADLINK gpib-32.dll.

    This class is used internally only.

    @private
    """

    def __init__(self):
        # load gpib dll
        gpibDll = ctypes.WinDLL("gpib-32.dll")

        # gpib init functions
        self._gpibIbfind = gpibDll.ibfindW
        self._gpibIbfind.restype = c_int
        self._gpibIbfind.argtypes = [c_wchar_p]

        self._gpibIbrsc = gpibDll.ibrsc
        self._gpibIbrsc.restype = c_int
        self._gpibIbrsc.argtypes = [c_int, c_int]

        self._gpibIbsic = gpibDll.ibsic
        self._gpibIbsic.restype = c_int
        self._gpibIbsic.argtypes = [c_int]

        # gpib send
        self._gpibSendImpl = gpibDll.Send
        self._gpibSendImpl.argtypes = [c_int, c_ushort, c_char_p, c_int, c_int]

        # gpib receive
        self._gpibReceiveImpl = gpibDll.Receive
        self._gpibReceiveImpl.argtypes = [c_int, c_ushort, c_void_p, c_int, c_int]

        # gpib receive
        self._gpibGetGlobalsImpl = gpibDll.gpib_get_globals
        self._gpibGetGlobalsImpl.restype = c_int
        self._gpibGetGlobalsImpl.argtypes = [c_void_p, c_void_p, c_void_p, c_void_p]

    def connect(self, board_name: str, address: int):
        # init
        self._board = self._gpibIbfind(board_name)  # "GPIB0")
        if self._board < 0:
            raise Exception(f"Board {board_name} does not exist!")

        self._gpibIbrsc(self._board, 1)
        self._gpibIbsic(self._board)

        self._addr = address
        self._max_len = 1000

    def get_globals(self):
        ibsta = c_int()
        iberr = c_int()
        ibcnt = c_int()
        ibcntl = c_int()
        self._gpibGetGlobalsImpl(
            byref(ibsta), byref(iberr), byref(ibcnt), byref(ibcntl)
        )
        return ibsta.value, iberr.value, ibcnt.value, ibcntl.value

    def send(self, str):
        self._gpibSendImpl(self._board, self._addr, str.encode("utf-8"), len(str), 1)

    def receive(self):
        byteArr = (c_byte * self._max_len)()
        self._gpibReceiveImpl(self._board, self._addr, byteArr, self._max_len, 0)
        ibsta, iberr, ibcnt, ibcntl = self.get_globals()
        return bytearray(byteArr[:ibcntl]).decode("utf-8")



from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase


class ModuleCommandGroupBase(CommandGroupBase):
    """Base class for all command groups."""

    def __init__(self, parent : 'SentioProber', abbr) -> None:
        super().__init__(parent)
        self._groupAbbr = abbr


    def get_prop(self, prop_name: str, arg1=None):
        """Query a module property.

        SENTIO exposes some of its internal variables in the form of so called "module properties".
        Module properties represent values that are identified by their name and can be of different
        types.

        A module property can also have an additional parameter.

        The remote command specification contains a list of some of exposed module properties.

        :param prop_name: The name of the property to query.
        :param arg1: An optional parameter for the property.
        :raises: ProberException if an error occured.
        """
        if arg1 == None:
            self.comm.send("{0}:get_prop {1}".format(self._groupAbbr, prop_name))
        else:
            self.comm.send("{0}:get_prop {1}, {2}".format(self._groupAbbr, prop_name, arg1))

        resp = Response.check_resp(self.comm.read_line())

        values = resp.message().split(",")

        # Try to figure out the type of the return value
        if len(values) == 1:
            # Test 1: Check if the returned value is a boolean. Do not use bool(...) because this
            #         would also convert numeric values
            #
            if resp.message().lower() == "true":
                return True
            elif resp.message().lower() == "false":
                return False

            # Test 2: Try to convert the return value into a number
            try:
                # I cannot distinguish between int and float reliably. Values will always be returned as floats
                return float(resp.message())
            except:
                return resp.message()
        elif len(values) == 2:
            return float(values[0]), float(values[1])
        elif len(values) == 3:
            return float(values[0]), float(values[1]), float(values[2])

        return int(resp.message())


    def set_prop(self, prop_name: str, *argv):
        """Set a module property.

        SENTIO exposes some of its internal variables in the form of so called "module properties".
        Module properties represent values that are identified by their name and can be of different
        types.

        A module property can also have an additional parameter.

        The remote command specification contains a list of some of exposed module properties.

        :param prop_name: The name of the property to query.
        :param argv: Optional parameters for the property.
        :raises: ProberException if an error occured.
        """
        cmd: str = self._groupAbbr + ":set_prop {0}"
        for n in range(0, len(argv)):
            cmd += ", {0}".format(argv[n])

        self.comm.send(cmd.format(prop_name))
        Response.check_resp(self.comm.read_line())
